__all__ = [
    "DiagramChart",
    "Folk54GSMDiagramChart",
    "Folk54SSCDiagramChart",
    "BP12GSMDiagramChart",
    "BP12SSCDiagramChart"]

import typing

import numpy as np

from ..model import GrainSizeSample
from ..statistic import get_GSM_proportion, get_SSC_proportion
from .BaseChart import BaseChart
from .config_matplotlib import normal_color


class DiagramChart(BaseChart):
    def __init__(self, parent=None, figsize=(8, 6)):
        super().__init__(parent=parent, figsize=figsize)
        self.axes = self.figure.subplots()
        self.draw_base()
        self.__sample_batchses = []

    def trans_pos(self, a, b):
        pass

    @property
    def title(self):
        pass

    @property
    def lines(self):
        pass

    @property
    def labels(self):
        pass

    def plot_legend(self):
        pass

    def draw_base(self):
        self.axes.axis("off")
        self.axes.set_aspect(1)
        self.axes.set_title(self.title)

        for sand, clay, kwargs in self.lines:
            x, y = self.trans_pos(sand, clay)
            self.axes.plot(x, y, **kwargs)

        for (sand, clay), text, kwargs in self.labels:
            x, y = self.trans_pos(sand, clay)
            self.axes.text(x, y, text, color=normal_color(), **kwargs)

        self.plot_legend()
        self.figure.tight_layout()

    def convert_samples(self, samples: typing.List[GrainSizeSample]):
        pass

    def show_samples(self, samples: typing.List[GrainSizeSample],
                     append=False, c="#ffffff00",
                     marker=".", ms=8, mfc="red", mew=0.0,
                     **kwargs):
        if len(samples) == 0:
            return
        if not append:
            self.axes.clear()
            self.draw_base()
            self.__sample_batchses.clear()
        a, b = self.convert_samples(samples)
        x, y = self.trans_pos(a, b)
        self.axes.plot(x, y, c=c, marker=marker, ms=ms, mfc=mfc, mew=mew, **kwargs)
        self.canvas.draw()
        plot_kwargs = dict(c=c, marker=marker, ms=ms, mfc=mfc, mew=mew)
        plot_kwargs.update(kwargs)
        self.__sample_batchses.append((samples, plot_kwargs))

    def update_chart(self):
        self.figure.clear()
        self.axes = self.figure.subplots()
        self.draw_base()
        for samples, plot_kwargs in self.__sample_batchses:
            a, b = self.convert_samples(samples)
            x, y = self.trans_pos(a, b)
            self.axes.plot(x, y, **plot_kwargs)
        self.canvas.draw()


class Folk54GSMDiagramChart(DiagramChart):
    def __init__(self, parent=None, figsize=(8, 6)):
        super().__init__(parent=parent, figsize=figsize)
        self.setWindowTitle(self.tr("GSM Diagram (Folk, 1954)"))

    @property
    def title(self):
        return "Gravel-sand-mud diagram (Folk, 1954)"

    @property
    def lines(self):
        STRUCTURAL_LINES = [
            ([0.0, 0.0, 1.0, 0.0], [0.0, 1.0, 0.0, 0.0], dict(c=normal_color(), linewidth=0.8)), # the 3 sides of this equilateral triangle
            ([0.0, 1.0], [0.01, 0.01], dict(c=normal_color(), linewidth=0.8)), # gravel = Trace
            ([0.0, 1.0], [0.05, 0.05], dict(c=normal_color(), linewidth=0.8)), # gravel = 5%
            ([0.0, 1.0], [0.3, 0.3], dict(c=normal_color(), linewidth=0.8)), # gravel = 30%
            ([0.0, 1.0], [0.8, 0.8], dict(c=normal_color(), linewidth=0.8)), # gravel = 80%

            ([1/10, 1/10], [0.0, 0.05], dict(c=normal_color(), linewidth=0.8)), # sand: mud = 1:9
            ([1/2, 1/2], [0.0, 0.8], dict(c=normal_color(), linewidth=0.8)), # sand: mud = 1:1
            ([9/10, 9/10], [0.0, 0.8], dict(c=normal_color(), linewidth=0.8))] # sand: mud = 9:1
        ADDTIONAL_LINES = [
            ([0.05, -0.03], [0.02, 0.12], dict(c=normal_color(), linewidth=0.8)),
            ([0.3, 0.3], [0.02, 0.06], dict(c=normal_color(), linewidth=0.8)),
            ([0.7, 0.7], [0.02, 0.06], dict(c=normal_color(), linewidth=0.8)),
            ([0.95, 1.03], [0.02, 0.12], dict(c=normal_color(), linewidth=0.8)),
            ([0.95, 1.03], [0.02, 0.12], dict(c=normal_color(), linewidth=0.8)),
            ([0.95, 1.03], [0.2, 0.275], dict(c=normal_color(), linewidth=0.8)),
            ([0.95, 1.03], [0.5, 0.55], dict(c=normal_color(), linewidth=0.8))]

        return STRUCTURAL_LINES + ADDTIONAL_LINES

    @property
    def labels(self):
        s = 0.01
        LABELS = [((-s, -s), "Mud", dict(ha="right", va="top", fontweight="bold")),
                  ((0.5, 1+s), "Gravel", dict(ha="center", va="bottom", fontweight="bold")),
                  ((1+s, -s), "Sand", dict(ha="left", va="top", fontweight="bold")),
                  ((-5*s, 0.55), "Gravel %", dict(ha="right", va="center", fontweight="bold")),
                  ((0.5, -5*s), "Sand:Mud Ratio", dict(ha="center", va="top", fontweight="bold")),

                  ((0.05, -s), "Mud", dict(ha="center", va="top", fontsize="x-small")),
                  ((0.3, -s), "Sandy Mud", dict(ha="center", va="top", fontsize="x-small")),
                  ((0.7, -s), "Muddy Sand", dict(ha="center", va="top", fontsize="x-small")),
                  ((0.95, -s), "Sand", dict(ha="center", va="top", fontsize="x-small")),

                  ((-0.03, 0.12), "Slightly\nGravelly\nMud", dict(ha="right", va="bottom", fontsize="x-small")),
                  ((0.3, 0.08), "Slightly Gravelly Sandy Mud", dict(ha="center", va="bottom", fontsize="x-small")),
                  ((0.7, 0.08), "Slightly Gravelly Muddy Sand", dict(ha="center", va="bottom", fontsize="x-small")),
                  ((1.03, 0.12), "Slightly\nGravelly\nSand", dict(ha="left", va="bottom", fontsize="x-small")),

                  ((0.2, 0.2), "Gravelly Mud", dict(ha="center", va="center", fontsize="x-small")),
                  ((0.7, 0.2), "Gravelly Muddy Sand", dict(ha="center", va="center", fontsize="x-small")),
                  ((1.03, 0.28), "Gravelly\nSand", dict(ha="left", va="bottom", fontsize="x-small")),
                  ((0.2, 0.55), "Muddy\nGravel", dict(ha="center", va="center", fontsize="x-small")),
                  ((0.7, 0.55), "Muddy\nSandy\nGravel", dict(ha="center", va="center", fontsize="x-small")),
                  ((1.03, 0.55), "Sandy\nGravel", dict(ha="left", va="bottom", fontsize="x-small")),
                  ((0.5, 0.85), "Gravel", dict(ha="center", va="bottom", fontsize="x-small")),

                  ((-1*s, 0.01), "Trace", dict(ha="right", va="center", fontsize="x-small")),
                  ((-1*s, 0.05), "5%", dict(ha="right", va="center", fontsize="x-small")),
                  ((-3*s, 0.3), "30%", dict(ha="right", va="center", fontsize="x-small")),
                  ((-8*s, 0.8), "80%", dict(ha="right", va="center", fontsize="x-small")),

                  ((1/10, -s), "1:9", dict(ha="center", va="top", fontsize="x-small")),
                  ((1/2, -s), "1:1", dict(ha="center", va="top", fontsize="x-small")),
                  ((9/10, -s), "9:1", dict(ha="center", va="top", fontsize="x-small"))]

        return LABELS

    def trans_pos(self, sand, gravel):
        sand, gravel = np.array(sand), np.array(gravel)
        y = 0.5 * np.sqrt(3) * gravel
        # calculate the cross of two lines
        # first line: y = 0.5 * np.sqrt(3) * gravel
        # second line: the line cross the two points, (sand, 0) and (0.5, 0.5*np.sqrt(3))
        key = np.not_equal(sand, 0.5)
        x = np.full_like(y, np.nan)
        a = np.full_like(y, np.nan)
        b = np.full_like(y, np.nan)
        x[~key] = 0.5
        a[key] = 0.5*np.sqrt(3) / (0.5-sand[key])
        b[key] = -a[key] * sand[key]
        x[key] = (y[key] - b[key]) / a[key]
        assert not np.any(np.isnan(x))
        return x, y

    def convert_samples(self, samples: typing.List[GrainSizeSample]):
        sand = []
        gravel = []
        for i, sample in enumerate(samples):
            gravel_i, sand_i, mud_i = get_GSM_proportion(sample.classes_φ, sample.distribution)
            sand.append(sand_i)
            gravel.append(gravel_i)
        return sand, gravel

    def retranslate(self):
        super().retranslate()
        self.setWindowTitle(self.tr("GSM Diagram (Folk, 1954)"))


class Folk54SSCDiagramChart(DiagramChart):
    def __init__(self, parent=None, figsize=(8, 6)):
        super().__init__(parent=parent, figsize=figsize)
        self.setWindowTitle(self.tr("SSC Diagram (Folk, 1954)"))

    @property
    def title(self):
        return "Sand-silt-clay diagram (Folk, 1954)"

    @property
    def lines(self):
        STRUCTURAL_LINES = [
            ([0.0, 0.0, 1.0, 0.0], [0.0, 1.0, 0.0, 0.0], dict(c=normal_color(), linewidth=0.8)), # the 3 sides of this equilateral triangle
            ([0.0, 1.0], [0.1, 0.1], dict(c=normal_color(), linewidth=0.8)), # sand = 10%
            ([0.0, 1.0], [0.5, 0.5], dict(c=normal_color(), linewidth=0.8)), # sand = 50%
            ([0.0, 1.0], [0.9, 0.9], dict(c=normal_color(), linewidth=0.8)), # sand = 90%
            ([1/3, 1/3], [0.0, 0.9], dict(c=normal_color(), linewidth=0.8)), # clay: silt = 1:2
            ([2/3, 2/3], [0.0, 0.9], dict(c=normal_color(), linewidth=0.8))] # clay: silt = 2:1
        ADDTIONAL_LINES = []

        return STRUCTURAL_LINES + ADDTIONAL_LINES

    @property
    def labels(self):
        s = 0.01
        LABELS = [
            ((-s, -s), "Clay", dict(ha="right", va="top", fontweight="bold")),
            ((0.5, 1+s), "Sand", dict(ha="center", va="bottom", fontweight="bold")),
            ((1+s, -s), "Silt", dict(ha="left", va="top", fontweight="bold")),
            ((-5*s, 0.55), "Sand %", dict(ha="right", va="center", fontweight="bold")),
            ((0.5, -3*s), "Silt:Clay Ratio", dict(ha="center", va="top", fontweight="bold")),

            ((0.16, 0.05), "Clay", dict(ha="center", va="center", fontsize="x-small")),
            ((0.5, 0.05), "Mud", dict(ha="center", va="center", fontsize="x-small")),
            ((0.84, 0.05), "Silt", dict(ha="center", va="center", fontsize="x-small")),
            ((0.16, 0.3), "Sandy Clay", dict(ha="center", va="center", fontsize="x-small")),
            ((0.5, 0.3), "Sandy Mud", dict(ha="center", va="center", fontsize="x-small")),
            ((0.84, 0.3), "Sandy Silt", dict(ha="center", va="center", fontsize="x-small")),
            ((0.16, 0.55), "Clayey Sand", dict(ha="center", va="center", fontsize="x-small")),
            ((0.5, 0.55), "Muddy Sand", dict(ha="center", va="center", fontsize="x-small")),
            ((0.84, 0.55), "Silty Sand", dict(ha="center", va="center", fontsize="x-small")),
            ((0.5, 0.91), "Sand", dict(ha="center", va="bottom", fontsize="x-small")),

            ((-1*s, 0.1), "10%", dict(ha="right", va="center", fontsize="x-small")),
            ((-5*s, 0.5), "50%", dict(ha="right", va="center", fontsize="x-small")),
            ((-9*s, 0.9), "90%", dict(ha="right", va="center", fontsize="x-small")),

            ((1/3, -s), "1:2", dict(ha="center", va="top", fontsize="x-small")),
            ((2/3, -s), "2:1", dict(ha="center", va="top", fontsize="x-small"))]

        return LABELS

    def trans_pos(self, silt, sand):
        silt, sand = np.array(silt), np.array(sand)
        y = 0.5 * np.sqrt(3) * sand
        # calculate the cross of two lines
        # first line: 0.5 * np.sqrt(3) * sand
        # second line: the line cross the two points, (clay, 0) and (0.5, 0.5*np.sqrt(3))
        # y = ax + b
        key = np.not_equal(silt, 0.5)
        x = np.full_like(y, np.nan)
        a = np.full_like(y, np.nan)
        b = np.full_like(y, np.nan)
        x[~key] = 0.5
        a[key] = 0.5*np.sqrt(3) / (0.5-silt[key])
        b[key] = -a[key] * silt[key]
        x[key] = (y[key] - b[key]) / a[key]
        assert not np.any(np.isnan(x))
        return x, y

    def convert_samples(self, samples: typing.List[GrainSizeSample]):
        silt = []
        sand = []
        for i, sample in enumerate(samples):
            sand_i, silt_i, clay_i = get_SSC_proportion(sample.classes_φ, sample.distribution)
            silt.append(silt_i)
            sand.append(sand_i)
        return silt, sand

    def retranslate(self):
        super().retranslate()
        self.setWindowTitle(self.tr("SSC Diagram (Folk, 1954)"))


class BP12GSMDiagramChart(DiagramChart):
    def __init__(self, parent=None, figsize=(8, 6)):
        super().__init__(parent=parent, figsize=figsize)
        self.setWindowTitle(self.tr("GSM Diagram (Blott & Pye, 2012)"))

    @property
    def title(self):
        return "Gravel-sand-mud diagram (Blott & Pye, 2012)"

    @property
    def lines(self):
        span = 0.01
        STRUCTURAL_LINES = [
            ([0.0, 0.0, 1.0, 0.0], [0.0, 1.0, 0.0, 0.0], dict(c=normal_color(), linewidth=0.8)), # the 3 sides of this equilateral triangle
            # Gravel
            ([0.0, 0.99], [0.01, 0.01], dict(c=normal_color(), linewidth=0.8)), # gravel = Trace
            ([0.0, 0.95], [0.05, 0.05], dict(c=normal_color(), linewidth=0.8)), # gravel = 5%
            ([0.0, 0.8], [0.2, 0.2], dict(c=normal_color(), linewidth=0.8)), # gravel = 20%
            ([0.0, 1/3, 0.5], [0.5, 1/3, 0.5], dict(c=normal_color(), linewidth=0.8)), # gravel = 50%, 33%
            # Sand
            ([0.01, 0.01], [0.0, 0.99], dict(c=normal_color(), linewidth=0.8)), # sand = Trace
            ([0.05, 0.05], [0.0, 0.95], dict(c=normal_color(), linewidth=0.8)), # sand = 5%
            ([0.2, 0.2], [0.0, 0.8], dict(c=normal_color(), linewidth=0.8)), # sand = 20%
            ([0.5, 1/3, 0.5], [0.0, 1/3, 0.5], dict(c=normal_color(), linewidth=0.8)), # sand = 50%, 33%
            # Mud
            ([0.99, 0.0], [0.0, 0.99], dict(c=normal_color(), linewidth=0.8)), # mud = Trace
            ([0.95, 0.0], [0.0, 0.95], dict(c=normal_color(), linewidth=0.8)), # mud = 5%
            ([0.8, 0.0], [0.0, 0.8], dict(c=normal_color(), linewidth=0.8)), # mud = 20%
            ([0.5, 1/3, 0.0], [0.0, 1/3, 0.5], dict(c=normal_color(), linewidth=0.8))] # mud = 50%, 33%
        ADDITIONAL_LINES = [
            ([0.0, -span], [0.0, -span], dict(c=normal_color(), linewidth=0.4)), # M
            ([0.0, -2*span], [1.0, 1+4*span], dict(c=normal_color(), linewidth=0.4)), # G
            ([1.0, 1+2*span], [0.0, -span], dict(c=normal_color(), linewidth=0.4)), # S
            # ticks of Gravel
            ([0.0, -span], [1.0, 1.0+span], dict(c=normal_color(), linewidth=0.4)),
            ([0.0, -span], [0.9, 0.9+span], dict(c=normal_color(), linewidth=0.4)),
            ([0.0, -span], [0.8, 0.8+span], dict(c=normal_color(), linewidth=0.4)),
            ([0.0, -span], [0.7, 0.7+span], dict(c=normal_color(), linewidth=0.4)),
            ([0.0, -span], [0.6, 0.6+span], dict(c=normal_color(), linewidth=0.4)),
            ([0.0, -span], [0.5, 0.5+span], dict(c=normal_color(), linewidth=0.4)),
            ([0.0, -span], [0.4, 0.4+span], dict(c=normal_color(), linewidth=0.4)),
            ([0.0, -span], [0.3, 0.3+span], dict(c=normal_color(), linewidth=0.4)),
            ([0.0, -span], [0.2, 0.2+span], dict(c=normal_color(), linewidth=0.4)),
            ([0.0, -span], [0.1, 0.1+span], dict(c=normal_color(), linewidth=0.4)),
            ([0.0, -span], [0.0, span], dict(c=normal_color(), linewidth=0.4)),
            # ticks of Sand
            ([0.0, 0.0+0.1*span], [1.0, 1.0+span], dict(c=normal_color(), linewidth=0.4)),
            ([0.1, 0.1+0.1*span], [0.9, 0.9+span], dict(c=normal_color(), linewidth=0.4)),
            ([0.2, 0.2+0.1*span], [0.8, 0.8+span], dict(c=normal_color(), linewidth=0.4)),
            ([0.3, 0.3+0.1*span], [0.7, 0.7+span], dict(c=normal_color(), linewidth=0.4)),
            ([0.4, 0.4+0.1*span], [0.6, 0.6+span], dict(c=normal_color(), linewidth=0.4)),
            ([0.5, 0.5+0.1*span], [0.5, 0.5+span], dict(c=normal_color(), linewidth=0.4)),
            ([0.6, 0.6+0.1*span], [0.4, 0.4+span], dict(c=normal_color(), linewidth=0.4)),
            ([0.7, 0.7+0.1*span], [0.3, 0.3+span], dict(c=normal_color(), linewidth=0.4)),
            ([0.8, 0.8+0.1*span], [0.2, 0.2+span], dict(c=normal_color(), linewidth=0.4)),
            ([0.9, 0.9+0.1*span], [0.1, 0.1+span], dict(c=normal_color(), linewidth=0.4)),
            ([1.0, 1.0+0.1*span], [0.0, span], dict(c=normal_color(), linewidth=0.4)),
            # ticks of Mud
            ([0.0, span], [0.0, -span], dict(c=normal_color(), linewidth=0.4)),
            ([0.1, 0.1+span], [0.0, -span], dict(c=normal_color(), linewidth=0.4)),
            ([0.2, 0.2+span], [0.0, -span], dict(c=normal_color(), linewidth=0.4)),
            ([0.3, 0.3+span], [0.0, -span], dict(c=normal_color(), linewidth=0.4)),
            ([0.4, 0.4+span], [0.0, -span], dict(c=normal_color(), linewidth=0.4)),
            ([0.5, 0.5+span], [0.0, -span], dict(c=normal_color(), linewidth=0.4)),
            ([0.6, 0.6+span], [0.0, -span], dict(c=normal_color(), linewidth=0.4)),
            ([0.7, 0.7+span], [0.0, -span], dict(c=normal_color(), linewidth=0.4)),
            ([0.8, 0.8+span], [0.0, -span], dict(c=normal_color(), linewidth=0.4)),
            ([0.9, 0.9+span], [0.0, -span], dict(c=normal_color(), linewidth=0.4)),
            ([1.0, 1.0-span], [0.0, -span], dict(c=normal_color(), linewidth=0.4)),

            # guide lines
            ([0.03, 0.03+4*span], [0.0, -2*span], dict(c=normal_color(), linewidth=0.4)),
            ([0.13, 0.13+4*span], [0.0, -2*span], dict(c=normal_color(), linewidth=0.4)),
            ([0.33, 0.33+4*span], [0.0, -2*span], dict(c=normal_color(), linewidth=0.4)),
            ([0.63, 0.63+4*span], [0.0, -2*span], dict(c=normal_color(), linewidth=0.4)),
            ([0.83, 0.83+4*span], [0.0, -2*span], dict(c=normal_color(), linewidth=0.4)),
            ([0.97, 0.97], [0.0, -2*span], dict(c=normal_color(), linewidth=0.4)),

            ([0.0, -2*span], [0.03, 0.03+2*span], dict(c=normal_color(), linewidth=0.4)),
            ([0.0, -2*span], [0.13, 0.13+2*span], dict(c=normal_color(), linewidth=0.4)),
            ([0.0, -2*span], [0.33, 0.33+2*span], dict(c=normal_color(), linewidth=0.4)),
            ([0.0, -2*span], [0.63, 0.63+2*span], dict(c=normal_color(), linewidth=0.4)),
            ([0.0, -2*span], [0.83, 0.83+2*span], dict(c=normal_color(), linewidth=0.4)),
            ([0.0, -2*span], [0.97, 0.97], dict(c=normal_color(), linewidth=0.4)),

            ([0.97, 0.97+0.2*span], [0.03, 0.03+2*span], dict(c=normal_color(), linewidth=0.4)),
            ([0.87, 0.87+0.2*span], [0.13, 0.13+2*span], dict(c=normal_color(), linewidth=0.4)),
            ([0.67, 0.67+0.2*span], [0.33, 0.33+2*span], dict(c=normal_color(), linewidth=0.4)),
            ([0.37, 0.37+0.2*span], [0.63, 0.63+2*span], dict(c=normal_color(), linewidth=0.4)),
            ([0.17, 0.17+0.2*span], [0.83, 0.83+2*span], dict(c=normal_color(), linewidth=0.4)),
            ([0.03, 0.03+0.2*span], [0.97, 0.97+2*span], dict(c=normal_color(), linewidth=0.4)),

            ([0.03, 0.03+2*span], [0.95, 0.95+2*span], dict(c=normal_color(), linewidth=0.4)),
            ([0.03, -2*span], [0.03, 0.03+4*span], dict(c=normal_color(), linewidth=0.4)),
            ([0.94, 0.94+0.4*span], [0.03, 0.03+4*span], dict(c=normal_color(), linewidth=0.4)),

            ]
        return STRUCTURAL_LINES + ADDITIONAL_LINES

    @property
    def labels(self):
        s = 0.01
        LABELS = [
            ((-s, -s), "M", dict(ha="right", va="top", fontweight="bold")),
            ((-2*s, 1+4*s), "G", dict(ha="center", va="bottom", fontweight="bold")),
            ((1+2*s, -s), "S", dict(ha="left", va="top", fontweight="bold")),
            ((-5*s, 0.55), "Gravel %", dict(ha="right", va="center", fontweight="bold")),
            ((0.5, 0.55), "Sand %", dict(ha="left", va="center", fontweight="bold")),
            ((0.5+5*s, -5*np.sqrt(3)*s), "Mud %", dict(ha="center", va="top", fontweight="bold")),
            # tick labels of Gravel
            ((-s, 1.0+s), "100%", dict(ha="right", va="bottom", fontsize="x-small")),
            ((-s, 0.9+s), "90%", dict(ha="right", va="bottom", fontsize="x-small")),
            ((-s, 0.8+s), "80%", dict(ha="right", va="bottom", fontsize="x-small")),
            ((-s, 0.7+s), "70%", dict(ha="right", va="bottom", fontsize="x-small")),
            ((-s, 0.6+s), "60%", dict(ha="right", va="bottom", fontsize="x-small")),
            ((-s, 0.5+s), "50%", dict(ha="right", va="bottom", fontsize="x-small")),
            ((-s, 0.4+s), "40%", dict(ha="right", va="bottom", fontsize="x-small")),
            ((-s, 0.3+s), "30%", dict(ha="right", va="bottom", fontsize="x-small")),
            ((-s, 0.2+s), "20%", dict(ha="right", va="bottom", fontsize="x-small")),
            ((-s, 0.1+s), "10%", dict(ha="right", va="bottom", fontsize="x-small")),
            ((-s, 0.0+s), "0%", dict(ha="right", va="bottom", fontsize="x-small")),
            # tick labels of Sand
            ((0.0, 1.0+s), "0%", dict(ha="left", va="bottom", fontsize="x-small")),
            ((0.1, 0.9+s), "10%", dict(ha="left", va="bottom", fontsize="x-small")),
            ((0.2, 0.8+s), "20%", dict(ha="left", va="bottom", fontsize="x-small")),
            ((0.3, 0.7+s), "30%", dict(ha="left", va="bottom", fontsize="x-small")),
            ((0.4, 0.6+s), "40%", dict(ha="left", va="bottom", fontsize="x-small")),
            ((0.5, 0.5+s), "50%", dict(ha="left", va="bottom", fontsize="x-small")),
            ((0.6, 0.4+s), "60%", dict(ha="left", va="bottom", fontsize="x-small")),
            ((0.7, 0.3+s), "70%", dict(ha="left", va="bottom", fontsize="x-small")),
            ((0.8, 0.2+s), "80%", dict(ha="left", va="bottom", fontsize="x-small")),
            ((0.9, 0.1+s), "90%", dict(ha="left", va="bottom", fontsize="x-small")),
            ((1.0, 0.0+s), "100%", dict(ha="left", va="bottom", fontsize="x-small")),
            # tick labels of Mud
            ((2*s, -s), "100%", dict(ha="center", va="top", fontsize="x-small", rotation=-45)),
            ((0.1+2*s, -s), "90%", dict(ha="center", va="top", fontsize="x-small", rotation=-45)),
            ((0.2+2*s, -s), "80%", dict(ha="center", va="top", fontsize="x-small", rotation=-45)),
            ((0.3+2*s, -s), "70%", dict(ha="center", va="top", fontsize="x-small", rotation=-45)),
            ((0.4+2*s, -s), "60%", dict(ha="center", va="top", fontsize="x-small", rotation=-45)),
            ((0.5+2*s, -s), "50%", dict(ha="center", va="top", fontsize="x-small", rotation=-45)),
            ((0.6+2*s, -s), "40%", dict(ha="center", va="top", fontsize="x-small", rotation=-45)),
            ((0.7+2*s, -s), "30%", dict(ha="center", va="top", fontsize="x-small", rotation=-45)),
            ((0.8+2*s, -s), "20%", dict(ha="center", va="top", fontsize="x-small", rotation=-45)),
            ((0.9+2*s, -s), "10%", dict(ha="center", va="top", fontsize="x-small", rotation=-45)),
            ((1.0-s, -s), "0%", dict(ha="center", va="top", fontsize="x-small", rotation=-45)),

            ((0.125, 0.025), "(vg)(s)M", dict(ha="center", va="center", fontsize="small")),
            ((0.35, 0.025), "(vg)sM", dict(ha="center", va="center", fontsize="small")),
            ((0.625, 0.025), "(vg)mS", dict(ha="center", va="center", fontsize="small")),
            ((0.850, 0.025), "(vg)(m)S", dict(ha="center", va="center", fontsize="small")),
            ((0.125, 0.125), "(g)(s)M", dict(ha="center", va="center", fontsize="small")),
            ((0.3, 0.125), "(g)sM", dict(ha="center", va="center", fontsize="small")),
            ((0.575, 0.125), "(g)mS", dict(ha="center", va="center", fontsize="small")),
            ((0.750, 0.125), "(g)(m)S", dict(ha="center", va="center", fontsize="small")),
            ((0.125, 0.35), "(s)gM", dict(ha="center", va="center", fontsize="small")),
            ((0.3, 0.3), "gsM", dict(ha="center", va="center", fontsize="small")),
            ((0.4, 0.3), "gmS", dict(ha="center", va="center", fontsize="small")),
            ((0.525, 0.35), "(m)gS", dict(ha="center", va="center", fontsize="small")),
            ((0.125, 0.55), "(s)mG", dict(ha="center", va="center", fontsize="small")),
            ((1/3-(0.45-1/3)/2, 0.45), "smG", dict(ha="center", va="center", fontsize="small")),
            ((0.325, 0.55), "(m)sG", dict(ha="center", va="center", fontsize="small")),
            ((0.125, 0.75), "(s)(m)G", dict(ha="center", va="center", fontsize="small")),

            ((0.03, 0.125), "(vs)(g)M", dict(ha="center", va="center", fontsize="small", rotation=60)),
            ((0.03, 0.35), "(vs)gM", dict(ha="center", va="center", fontsize="small", rotation=60)),
            ((0.03, 0.6), "(vs)mG", dict(ha="center", va="center", fontsize="small", rotation=60)),
            ((0.03, 0.85), "(vs)(m)G", dict(ha="center", va="center", fontsize="small", rotation=60)),

            ((0.845, 0.125), "(vm)(g)S", dict(ha="center", va="center", fontsize="small", rotation=-60)),
            ((0.62, 0.35), "(vm)gS", dict(ha="center", va="center", fontsize="small", rotation=-60)),
            ((0.37, 0.6), "(vm)sG", dict(ha="center", va="center", fontsize="small", rotation=-60)),
            ((0.12, 0.85), "(vm)(s)G", dict(ha="center", va="center", fontsize="small", rotation=-60)),

            ((0.05+4*s, -2*s), "(vs)M", dict(ha="center", va="top", fontsize="small", rotation=-45)),
            ((0.14+4*s, -2*s), "(s)M", dict(ha="center", va="top", fontsize="small")),
            ((0.34+4*s, -2*s), "sM", dict(ha="center", va="top", fontsize="small")),
            ((0.64+4*s, -2*s), "mS", dict(ha="center", va="top", fontsize="small")),
            ((0.84+4*s, -2*s), "(m)S", dict(ha="center", va="top", fontsize="small")),
            ((0.94+4*s, -2*s), "(vm)S", dict(ha="center", va="top", fontsize="small", rotation=-45)),

            ((-2*s, 0.03+2*s), "(vg)M", dict(ha="right", va="center", fontsize="small")),
            ((-2*s, 0.13+2*s), "(g)M", dict(ha="right", va="bottom", fontsize="small")),
            ((-2*s, 0.33+2*s), "gM", dict(ha="right", va="bottom", fontsize="small")),
            ((-2*s, 0.63+2*s), "mG", dict(ha="right", va="bottom", fontsize="small")),
            ((-2*s, 0.83+2*s), "(m)G", dict(ha="right", va="bottom", fontsize="small")),
            ((-2*s, 0.97), "(vm)G", dict(ha="right", va="center", fontsize="small")),

            ((0.97+0.2*s, 0.03+2*s), "(vg)S", dict(ha="left", va="center", fontsize="small")),
            ((0.87+0.2*s, 0.13+2*s), "(g)S", dict(ha="left", va="bottom", fontsize="small")),
            ((0.67+0.2*s, 0.33+2*s), "gS", dict(ha="left", va="bottom", fontsize="small")),
            ((0.37+0.2*s, 0.63+2*s), "sG", dict(ha="left", va="bottom", fontsize="small")),
            ((0.17+0.2*s, 0.83+2*s), "(s)G", dict(ha="left", va="bottom", fontsize="small")),
            ((0.03+0.2*s, 0.97+2*s), "(vs)G", dict(ha="left", va="bottom", fontsize="small")),

            ((0.03+0.5*s, 0.97+2*s), "(vs)(vm)G", dict(ha="left", va="top", fontsize="small")),
            ((-2*s, 0.03+4*s), "(vg)(vs)G", dict(ha="right", va="bottom", fontsize="small")),
            ((0.94+0.4*s, 0.03+4*s), "(vg)(vm)G", dict(ha="left", va="bottom", fontsize="small"))]

        return LABELS

    def plot_legend(self):
        x0, y0 = 0.7, 0.9
        w, h = 0.5, 0.16
        span = 0.01
        row_h = 0.02
        self.axes.plot([x0, x0+w, x0+w, x0, x0], [y0, y0, y0-h, y0-h, y0], c=normal_color(), linewidth=0.8)
        self.axes.text(x0+w/2, y0-span, "Conventions", ha="center", va="top", fontsize="small", color=normal_color())
        texts = ["UPPER CASE - Largest component (noun)",
                 "Lower case - Descriptive term (adjective)",
                 "( ) - Slightly (qualification)",
                 "(v ) - Very slightly (qualification)"]
        for row, text in enumerate(texts, 1):
            x = x0+2*span
            y = y0-span - row*(row_h+span)
            self.axes.text(x, y, text, ha="left", va="top", fontsize="x-small", color=normal_color())

        x0, y0 = 0.9, 0.7
        w, h = 0.3, 0.32
        span = 0.01
        row_h = 0.02
        self.axes.plot([x0, x0+w, x0+w, x0, x0], [y0, y0, y0-h, y0-h, y0], c=normal_color(), linewidth=0.8)
        self.axes.text(x0+w/2, y0-span, "Term used", ha="center", va="top", fontsize="small", color=normal_color())
        texts = ["G - Gravel    g - Gravelly",
                 "S - Sand      s - Sandy",
                 "M - Mud       m - Muddy",
                 "(g)  - Slightly gravelly",
                 "(s)  - Slightly sandy",
                 "(m)  - Slightly muddy",
                 "(vg) - Very slightly gravelly",
                 "(vs) - Very slightly sandy",
                 "(vm) - Very slightly muddy"]
        for row, text in enumerate(texts, 1):
            x = x0+2*span
            y = y0-span - row*(row_h+span)
            self.axes.text(x, y, text, ha="left", va="top", fontsize="x-small", color=normal_color())

    def trans_pos(self, a, b):
        a, b = np.array(a), np.array(b)
        y = 0.5 * np.sqrt(3) * b
        x = 0.5*b + a
        return x, y

    def convert_samples(self, samples: typing.List[GrainSizeSample]):
        sand = []
        gravel = []
        for i, sample in enumerate(samples):
            gravel_i, sand_i, mud_i = get_GSM_proportion(sample.classes_φ, sample.distribution)
            sand.append(sand_i)
            gravel.append(gravel_i)
        return sand, gravel

    def retranslate(self):
        super().retranslate()
        self.setWindowTitle(self.tr("GSM Diagram (Blott & Pye, 2012)"))


class BP12SSCDiagramChart(DiagramChart):
    def __init__(self, parent=None, figsize=(8, 6)):
        super().__init__(parent=parent, figsize=figsize)
        self.setWindowTitle(self.tr("SSC Diagram (Blott & Pye, 2012)"))

    @property
    def title(self):
        return "Sand-silt-clay diagram (Blott & Pye, 2012)"

    @property
    def lines(self):
        span = 0.01
        STRUCTURAL_LINES = [
            ([0.0, 0.0, 1.0, 0.0], [0.0, 1.0, 0.0, 0.0], dict(c=normal_color(), linewidth=0.8)), # the 3 sides of this equilateral triangle
            # Sand
            ([0.0, 0.99], [0.01, 0.01], dict(c=normal_color(), linewidth=0.8)), # sand = Trace
            ([0.0, 0.95], [0.05, 0.05], dict(c=normal_color(), linewidth=0.8)), # sand = 5%
            ([0.0, 0.8], [0.2, 0.2], dict(c=normal_color(), linewidth=0.8)), # sand = 20%
            ([0.0, 1/3, 0.5], [0.5, 1/3, 0.5], dict(c=normal_color(), linewidth=0.8)), # sand = 50%, 33%
            # Silt
            ([0.01, 0.01], [0.0, 0.99], dict(c=normal_color(), linewidth=0.8)), # silt = Trace
            ([0.05, 0.05], [0.0, 0.95], dict(c=normal_color(), linewidth=0.8)), # silt = 5%
            ([0.2, 0.2], [0.0, 0.8], dict(c=normal_color(), linewidth=0.8)), # silt = 20%
            ([0.5, 1/3, 0.5], [0.0, 1/3, 0.5], dict(c=normal_color(), linewidth=0.8)), # silt = 50%, 33%
            # Clay
            ([0.99, 0.0], [0.0, 0.99], dict(c=normal_color(), linewidth=0.8)), # clay = Trace
            ([0.95, 0.0], [0.0, 0.95], dict(c=normal_color(), linewidth=0.8)), # clay = 5%
            ([0.8, 0.0], [0.0, 0.8], dict(c=normal_color(), linewidth=0.8)), # clay = 20%
            ([0.5, 1/3, 0.0], [0.0, 1/3, 0.5], dict(c=normal_color(), linewidth=0.8))] # clay = 50%, 33%
        ADDITIONAL_LINES = [
            ([0.0, -span], [0.0, -span], dict(c=normal_color(), linewidth=0.4)), # S
            ([0.0, -2*span], [1.0, 1+4*span], dict(c=normal_color(), linewidth=0.4)), # SI
            ([1.0, 1+2*span], [0.0, -span], dict(c=normal_color(), linewidth=0.4)), # C
            # ticks of Sand
            ([0.0, -span], [1.0, 1.0+span], dict(c=normal_color(), linewidth=0.4)),
            ([0.0, -span], [0.9, 0.9+span], dict(c=normal_color(), linewidth=0.4)),
            ([0.0, -span], [0.8, 0.8+span], dict(c=normal_color(), linewidth=0.4)),
            ([0.0, -span], [0.7, 0.7+span], dict(c=normal_color(), linewidth=0.4)),
            ([0.0, -span], [0.6, 0.6+span], dict(c=normal_color(), linewidth=0.4)),
            ([0.0, -span], [0.5, 0.5+span], dict(c=normal_color(), linewidth=0.4)),
            ([0.0, -span], [0.4, 0.4+span], dict(c=normal_color(), linewidth=0.4)),
            ([0.0, -span], [0.3, 0.3+span], dict(c=normal_color(), linewidth=0.4)),
            ([0.0, -span], [0.2, 0.2+span], dict(c=normal_color(), linewidth=0.4)),
            ([0.0, -span], [0.1, 0.1+span], dict(c=normal_color(), linewidth=0.4)),
            ([0.0, -span], [0.0, span], dict(c=normal_color(), linewidth=0.4)),
            # ticks of Silt
            ([0.0, 0.0+0.1*span], [1.0, 1.0+span], dict(c=normal_color(), linewidth=0.4)),
            ([0.1, 0.1+0.1*span], [0.9, 0.9+span], dict(c=normal_color(), linewidth=0.4)),
            ([0.2, 0.2+0.1*span], [0.8, 0.8+span], dict(c=normal_color(), linewidth=0.4)),
            ([0.3, 0.3+0.1*span], [0.7, 0.7+span], dict(c=normal_color(), linewidth=0.4)),
            ([0.4, 0.4+0.1*span], [0.6, 0.6+span], dict(c=normal_color(), linewidth=0.4)),
            ([0.5, 0.5+0.1*span], [0.5, 0.5+span], dict(c=normal_color(), linewidth=0.4)),
            ([0.6, 0.6+0.1*span], [0.4, 0.4+span], dict(c=normal_color(), linewidth=0.4)),
            ([0.7, 0.7+0.1*span], [0.3, 0.3+span], dict(c=normal_color(), linewidth=0.4)),
            ([0.8, 0.8+0.1*span], [0.2, 0.2+span], dict(c=normal_color(), linewidth=0.4)),
            ([0.9, 0.9+0.1*span], [0.1, 0.1+span], dict(c=normal_color(), linewidth=0.4)),
            ([1.0, 1.0+0.1*span], [0.0, span], dict(c=normal_color(), linewidth=0.4)),
            # ticks of Caly
            ([0.0, span], [0.0, -span], dict(c=normal_color(), linewidth=0.4)),
            ([0.1, 0.1+span], [0.0, -span], dict(c=normal_color(), linewidth=0.4)),
            ([0.2, 0.2+span], [0.0, -span], dict(c=normal_color(), linewidth=0.4)),
            ([0.3, 0.3+span], [0.0, -span], dict(c=normal_color(), linewidth=0.4)),
            ([0.4, 0.4+span], [0.0, -span], dict(c=normal_color(), linewidth=0.4)),
            ([0.5, 0.5+span], [0.0, -span], dict(c=normal_color(), linewidth=0.4)),
            ([0.6, 0.6+span], [0.0, -span], dict(c=normal_color(), linewidth=0.4)),
            ([0.7, 0.7+span], [0.0, -span], dict(c=normal_color(), linewidth=0.4)),
            ([0.8, 0.8+span], [0.0, -span], dict(c=normal_color(), linewidth=0.4)),
            ([0.9, 0.9+span], [0.0, -span], dict(c=normal_color(), linewidth=0.4)),
            ([1.0, 1.0-span], [0.0, -span], dict(c=normal_color(), linewidth=0.4)),

            # guide lines
            ([0.03, 0.03+4*span], [0.0, -2*span], dict(c=normal_color(), linewidth=0.4)),
            ([0.13, 0.13+4*span], [0.0, -2*span], dict(c=normal_color(), linewidth=0.4)),
            ([0.33, 0.33+4*span], [0.0, -2*span], dict(c=normal_color(), linewidth=0.4)),
            ([0.63, 0.63+4*span], [0.0, -2*span], dict(c=normal_color(), linewidth=0.4)),
            ([0.83, 0.83+4*span], [0.0, -2*span], dict(c=normal_color(), linewidth=0.4)),
            ([0.97, 0.97], [0.0, -2*span], dict(c=normal_color(), linewidth=0.4)),

            ([0.0, -2*span], [0.03, 0.03+2*span], dict(c=normal_color(), linewidth=0.4)),
            ([0.0, -2*span], [0.13, 0.13+2*span], dict(c=normal_color(), linewidth=0.4)),
            ([0.0, -2*span], [0.33, 0.33+2*span], dict(c=normal_color(), linewidth=0.4)),
            ([0.0, -2*span], [0.63, 0.63+2*span], dict(c=normal_color(), linewidth=0.4)),
            ([0.0, -2*span], [0.83, 0.83+2*span], dict(c=normal_color(), linewidth=0.4)),
            ([0.0, -2*span], [0.97, 0.97], dict(c=normal_color(), linewidth=0.4)),

            ([0.97, 0.97+0.2*span], [0.03, 0.03+2*span], dict(c=normal_color(), linewidth=0.4)),
            ([0.87, 0.87+0.2*span], [0.13, 0.13+2*span], dict(c=normal_color(), linewidth=0.4)),
            ([0.67, 0.67+0.2*span], [0.33, 0.33+2*span], dict(c=normal_color(), linewidth=0.4)),
            ([0.37, 0.37+0.2*span], [0.63, 0.63+2*span], dict(c=normal_color(), linewidth=0.4)),
            ([0.17, 0.17+0.2*span], [0.83, 0.83+2*span], dict(c=normal_color(), linewidth=0.4)),
            ([0.03, 0.03+0.2*span], [0.97, 0.97+2*span], dict(c=normal_color(), linewidth=0.4)),

            ([0.03, 0.03+2*span], [0.95, 0.95+2*span], dict(c=normal_color(), linewidth=0.4)),
            ([0.03, -2*span], [0.03, 0.03+4*span], dict(c=normal_color(), linewidth=0.4)),
            ([0.94, 0.94+0.4*span], [0.03, 0.03+4*span], dict(c=normal_color(), linewidth=0.4)),

            ]
        return STRUCTURAL_LINES + ADDITIONAL_LINES

    @property
    def labels(self):
        s = 0.01
        LABELS = [
            ((-s, -s), "C", dict(ha="right", va="top", fontweight="bold")),
            ((-2*s, 1+4*s), "S", dict(ha="center", va="bottom", fontweight="bold")),
            ((1+2*s, -s), "SI", dict(ha="left", va="top", fontweight="bold")),
            ((-5*s, 0.55), "Sand %", dict(ha="right", va="center", fontweight="bold")),
            ((0.5, 0.55), "Silt %", dict(ha="left", va="center", fontweight="bold")),
            ((0.5+5*s, -5*np.sqrt(3)*s), "Clay %", dict(ha="center", va="top", fontweight="bold")),
            # tick labels of Sand
            ((-s, 1.0+s), "100%", dict(ha="right", va="bottom", fontsize="x-small")),
            ((-s, 0.9+s), "90%", dict(ha="right", va="bottom", fontsize="x-small")),
            ((-s, 0.8+s), "80%", dict(ha="right", va="bottom", fontsize="x-small")),
            ((-s, 0.7+s), "70%", dict(ha="right", va="bottom", fontsize="x-small")),
            ((-s, 0.6+s), "60%", dict(ha="right", va="bottom", fontsize="x-small")),
            ((-s, 0.5+s), "50%", dict(ha="right", va="bottom", fontsize="x-small")),
            ((-s, 0.4+s), "40%", dict(ha="right", va="bottom", fontsize="x-small")),
            ((-s, 0.3+s), "30%", dict(ha="right", va="bottom", fontsize="x-small")),
            ((-s, 0.2+s), "20%", dict(ha="right", va="bottom", fontsize="x-small")),
            ((-s, 0.1+s), "10%", dict(ha="right", va="bottom", fontsize="x-small")),
            ((-s, 0.0+s), "0%", dict(ha="right", va="bottom", fontsize="x-small")),
            # tick labels of Silt
            ((0.0, 1.0+s), "0%", dict(ha="left", va="bottom", fontsize="x-small")),
            ((0.1, 0.9+s), "10%", dict(ha="left", va="bottom", fontsize="x-small")),
            ((0.2, 0.8+s), "20%", dict(ha="left", va="bottom", fontsize="x-small")),
            ((0.3, 0.7+s), "30%", dict(ha="left", va="bottom", fontsize="x-small")),
            ((0.4, 0.6+s), "40%", dict(ha="left", va="bottom", fontsize="x-small")),
            ((0.5, 0.5+s), "50%", dict(ha="left", va="bottom", fontsize="x-small")),
            ((0.6, 0.4+s), "60%", dict(ha="left", va="bottom", fontsize="x-small")),
            ((0.7, 0.3+s), "70%", dict(ha="left", va="bottom", fontsize="x-small")),
            ((0.8, 0.2+s), "80%", dict(ha="left", va="bottom", fontsize="x-small")),
            ((0.9, 0.1+s), "90%", dict(ha="left", va="bottom", fontsize="x-small")),
            ((1.0, 0.0+s), "100%", dict(ha="left", va="bottom", fontsize="x-small")),
            # tick labels of Clay
            ((2*s, -s), "100%", dict(ha="center", va="top", fontsize="x-small", rotation=-45)),
            ((0.1+2*s, -s), "90%", dict(ha="center", va="top", fontsize="x-small", rotation=-45)),
            ((0.2+2*s, -s), "80%", dict(ha="center", va="top", fontsize="x-small", rotation=-45)),
            ((0.3+2*s, -s), "70%", dict(ha="center", va="top", fontsize="x-small", rotation=-45)),
            ((0.4+2*s, -s), "60%", dict(ha="center", va="top", fontsize="x-small", rotation=-45)),
            ((0.5+2*s, -s), "50%", dict(ha="center", va="top", fontsize="x-small", rotation=-45)),
            ((0.6+2*s, -s), "40%", dict(ha="center", va="top", fontsize="x-small", rotation=-45)),
            ((0.7+2*s, -s), "30%", dict(ha="center", va="top", fontsize="x-small", rotation=-45)),
            ((0.8+2*s, -s), "20%", dict(ha="center", va="top", fontsize="x-small", rotation=-45)),
            ((0.9+2*s, -s), "10%", dict(ha="center", va="top", fontsize="x-small", rotation=-45)),
            ((1.0-s, -s), "0%", dict(ha="center", va="top", fontsize="x-small", rotation=-45)),

            ((0.125, 0.025), "(vs)(si)C", dict(ha="center", va="center", fontsize="small")),
            ((0.35, 0.025), "(vs)siC", dict(ha="center", va="center", fontsize="small")),
            ((0.625, 0.025), "(vs)cSI", dict(ha="center", va="center", fontsize="small")),
            ((0.850, 0.025), "(vs)(c)SI", dict(ha="center", va="center", fontsize="small")),

            ((0.125, 0.125), "(s)(si)C", dict(ha="center", va="center", fontsize="small")),
            ((0.3, 0.125), "(s)siC", dict(ha="center", va="center", fontsize="small")),
            ((0.575, 0.125), "(s)cSI", dict(ha="center", va="center", fontsize="small")),
            ((0.750, 0.125), "(s)(c)SI", dict(ha="center", va="center", fontsize="small")),
            ((0.125, 0.35), "(si)sC", dict(ha="center", va="center", fontsize="small")),
            ((0.3, 0.3), "ssiC", dict(ha="center", va="center", fontsize="small")),
            ((0.4, 0.3), "scSI", dict(ha="center", va="center", fontsize="small")),
            ((0.525, 0.35), "(c)sSI", dict(ha="center", va="center", fontsize="small")),
            ((0.125, 0.55), "(si)cS", dict(ha="center", va="center", fontsize="small")),
            ((1/3-(0.45-1/3)/2, 0.45), "sicS", dict(ha="center", va="center", fontsize="small")),
            ((0.325, 0.55), "(c)siS", dict(ha="center", va="center", fontsize="small")),
            ((0.125, 0.75), "(si)(c)S", dict(ha="center", va="center", fontsize="small")),

            ((0.03, 0.125), "(vsi)(s)C", dict(ha="center", va="center", fontsize="small", rotation=60)),
            ((0.03, 0.35), "(vsi)sC", dict(ha="center", va="center", fontsize="small", rotation=60)),
            ((0.03, 0.6), "(vsi)cS", dict(ha="center", va="center", fontsize="small", rotation=60)),
            ((0.03, 0.85), "(vsi)(c)S", dict(ha="center", va="center", fontsize="small", rotation=60)),

            ((0.845, 0.125), "(vc)(s)SI", dict(ha="center", va="center", fontsize="small", rotation=-60)),
            ((0.62, 0.35), "(vc)sSI", dict(ha="center", va="center", fontsize="small", rotation=-60)),
            ((0.37, 0.6), "(vc)siS", dict(ha="center", va="center", fontsize="small", rotation=-60)),
            ((0.12, 0.85), "(vc)(si)S", dict(ha="center", va="center", fontsize="small", rotation=-60)),

            ((0.05+4*s, -2*s), "(vsi)C", dict(ha="center", va="top", fontsize="small", rotation=-45)),
            ((0.14+4*s, -2*s), "(si)C", dict(ha="center", va="top", fontsize="small")),
            ((0.34+4*s, -2*s), "siC", dict(ha="center", va="top", fontsize="small")),
            ((0.64+4*s, -2*s), "cSI", dict(ha="center", va="top", fontsize="small")),
            ((0.84+4*s, -2*s), "(c)SI", dict(ha="center", va="top", fontsize="small")),
            ((0.94+4*s, -2*s), "(vc)SI", dict(ha="center", va="top", fontsize="small", rotation=-45)),

            ((-2*s, 0.03+2*s), "(vs)C", dict(ha="right", va="center", fontsize="small")),
            ((-2*s, 0.13+2*s), "(s)C", dict(ha="right", va="bottom", fontsize="small")),
            ((-2*s, 0.33+2*s), "sC", dict(ha="right", va="bottom", fontsize="small")),
            ((-2*s, 0.63+2*s), "cS", dict(ha="right", va="bottom", fontsize="small")),
            ((-2*s, 0.83+2*s), "(c)S", dict(ha="right", va="bottom", fontsize="small")),
            ((-2*s, 0.97), "(vc)S", dict(ha="right", va="center", fontsize="small")),

            ((0.97+0.2*s, 0.03+2*s), "(vs)SI", dict(ha="left", va="center", fontsize="small")),
            ((0.87+0.2*s, 0.13+2*s), "(s)SI", dict(ha="left", va="bottom", fontsize="small")),
            ((0.67+0.2*s, 0.33+2*s), "sSI", dict(ha="left", va="bottom", fontsize="small")),
            ((0.37+0.2*s, 0.63+2*s), "siS", dict(ha="left", va="bottom", fontsize="small")),
            ((0.17+0.2*s, 0.83+2*s), "(si)S", dict(ha="left", va="bottom", fontsize="small")),
            ((0.03+0.2*s, 0.97+2*s), "(vsi)S", dict(ha="left", va="bottom", fontsize="small")),

            ((0.03+0.5*s, 0.97+2*s), "(vsi)(vc)S", dict(ha="left", va="top", fontsize="small")),
            ((-2*s, 0.03+4*s), "(vs)(vsi)C", dict(ha="right", va="bottom", fontsize="small")),
            ((0.94+0.4*s, 0.03+4*s), "(vs)(vc)SI", dict(ha="left", va="bottom", fontsize="small"))]

        return LABELS

    def trans_pos(self, a, b):
        a, b = np.array(a), np.array(b)
        y = 0.5 * np.sqrt(3) * b
        x = 0.5*b + a
        return x, y

    def plot_legend(self):
        x0, y0 = 0.7, 0.9
        w, h = 0.5, 0.16
        span = 0.01
        row_h = 0.02
        self.axes.plot([x0, x0+w, x0+w, x0, x0], [y0, y0, y0-h, y0-h, y0], c=normal_color(), linewidth=0.8)
        self.axes.text(x0+w/2, y0-span, "Conventions", ha="center", va="top", fontsize="small", color=normal_color())
        texts = ["UPPER CASE - Largest component (noun)",
                 "Lower case - Descriptive term (adjective)",
                 "( ) - Slightly (qualification)",
                 "(v ) - Very slightly (qualification)"]
        for row, text in enumerate(texts, 1):
            x = x0+2*span
            y = y0-span - row*(row_h+span)
            self.axes.text(x, y, text, ha="left", va="top", fontsize="x-small", color=normal_color())

        x0, y0 = 0.9, 0.7
        w, h = 0.3, 0.32
        span = 0.01
        row_h = 0.02
        self.axes.plot([x0, x0+w, x0+w, x0, x0], [y0, y0, y0-h, y0-h, y0], c=normal_color(), linewidth=0.8)
        self.axes.text(x0+w/2, y0-span, "Term used", ha="center", va="top", fontsize="small", color=normal_color())
        texts = ["S  - Sand     s  - Sandy",
                 "SI - Silt     si - Silty",
                 "C  - Clay     c  - Clayey",
                 "(s)   - Slightly sandy",
                 "(si)  - Slightly silty",
                 "(c)   - Slightly clayey",
                 "(vs)  - Very slightly sandy",
                 "(vsi) - Very slightly silty",
                 "(vc)  - Very slightly clayey"]
        for row, text in enumerate(texts, 1):
            x = x0+2*span
            y = y0-span - row*(row_h+span)
            self.axes.text(x, y, text, ha="left", va="top", fontsize="x-small", color=normal_color())

    def convert_samples(self, samples: typing.List[GrainSizeSample]):
        silt = []
        sand = []
        for i, sample in enumerate(samples):
            sand_i, silt_i, clay_i = get_SSC_proportion(sample.classes_φ, sample.distribution)
            silt.append(silt_i)
            sand.append(sand_i)

        return silt, sand

    def retranslate(self):
        super().retranslate()
        self.setWindowTitle(self.tr("SSC Diagram (Blott & Pye, 2012)"))



import numpy as np


def log10MSE_distance(values: np.ndarray, targets: np.ndarray, axis=None) -> float:
    return np.log10(np.mean(np.square(values - targets), axis=axis))

def MSE_distance(values: np.ndarray, targets: np.ndarray, axis=None) -> float:
    return np.mean(np.square(values - targets), axis=axis)

def p_norm(values: np.ndarray, targets: np.ndarray, p=2, axis=None) -> float:
    return np.sum(np.abs(values - targets) ** p, axis=axis) ** (1 / p)

def cosine_distance(values: np.ndarray, targets: np.ndarray, axis=None) -> float:
    if np.all(np.equal(values, 0.0)) or np.all(np.equal(targets, 0.0)):
        return 1.0
    cosine = np.sum(values * targets, axis=axis) / (np.sqrt(np.sum(np.square(values), axis=axis)) * np.sqrt(np.sum(np.square(targets), axis=axis)))
    return abs(cosine)

def angular_distance(values: np.ndarray, targets: np.ndarray, axis=None) -> float:
    cosine = cosine_distance(values, targets, axis=axis)
    angular = 2 * np.arccos(cosine) / np.pi
    return angular

def get_distance_function(distance: str):
    if distance[-4:] == "norm":
        p = int(distance[0])
        return lambda x, y, axis=None: p_norm(x, y, p, axis=axis)
    elif distance == "MSE":
        return lambda x, y, axis=None: MSE_distance(x, y, axis=axis)
    elif distance == "log10MSE":
        return lambda x, y, axis=None: log10MSE_distance(x, y, axis=axis)
    elif distance == "cosine":
        return lambda x, y, axis=None: cosine_distance(x, y, axis=axis)
    elif distance == "angular":
        return lambda x, y, axis=None: angular_distance(x, y, axis=axis)
    else:
        raise NotImplementedError(distance)



import typing

import numpy as np

from ..model import GrainSizeDataset, GrainSizeSample
from ..ssu import DISTRIBUTION_CLASS_MAP, DistributionType, SSUViewModel
from ..statistic import convert_μm_to_φ, convert_φ_to_μm


class ArtificialComponent:
    def __init__(self,
                 distribution: np.ndarray,
                 proportion: float,
                 moments: typing.Tuple[float, float, float, float]):
        self.__distribution = distribution
        self.__proportion = proportion
        m, v, s, k = moments
        self.__moments = dict(mean=m, std=np.sqrt(v), skewness=s, kurtosis=k)

    @property
    def distribution(self) -> np.ndarray:
        return self.__distribution

    @property
    def proportion(self) -> float:
        return self.__proportion

    @property
    def moments(self) -> dict:
        return self.__moments


class ArtificialSample:
    def __init__(self,
                 name: str,
                 classes_μm: np.ndarray,
                 classes_φ: np.ndarray,
                 distribution: np.ndarray,
                 components: typing.Iterable[np.ndarray],
                 proportions: typing.Iterable[float],
                 moments: typing.Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]):
        self.__name = name
        self.__classes_μm = classes_μm
        self.__classes_φ = classes_φ
        self.__distribution = distribution
        self.__components = [] # type: list[ArtificialComponent]
        m, v, s, k = moments
        for component, proportion, moments in zip(components, proportions, zip(m, v, s, k)):
            artificial_component = ArtificialComponent(component, proportion, moments)
            self.__components.append(artificial_component)
        self.__components.sort(key=lambda component: component.moments["mean"], reverse=True)

    @property
    def name(self) -> str:
        return self.__name

    @name.setter
    def name(self, value: str):
        assert isinstance(value, str)
        self.__name = value

    @property
    def classes_μm(self) -> np.ndarray:
        return self.__classes_μm

    @property
    def classes_φ(self) -> np.ndarray:
        return self.__classes_φ

    @property
    def distribution(self) -> np.ndarray:
        return self.__distribution

    @property
    def n_components(self) -> int:
        return len(self.__components)

    @property
    def components(self) -> typing.List[ArtificialComponent]:
        return self.__components

    @property
    def sample_to_fit(self):
        sample = GrainSizeSample(self.name, self.classes_μm, self.classes_φ, self.distribution)
        return sample

    @property
    def view_model(self):
        distributions = [component.distribution for component in self.components]
        proportions = [component.proportion for component in self.components]

        model = SSUViewModel(
            self.classes_φ, self.distribution,
            self.distribution, distributions, proportions,
            title=self.name, component_prefix="AC")
        return model


class ArtificialDataset:
    def __init__(self,
                 target: typing.Iterable[typing.Iterable[typing.Tuple[float, float]]],
                 distribution_type: DistributionType,
                 n_samples: int,
                 params: np.ndarray = None,
                 min_μm=0.02, max_μm=2000.0, n_classes=101,
                 precision=4, noise=5):
        # do validations
        if params is None:
            params = get_params(target, n_samples)
        else:
            assert isinstance(params, np.ndarray)
            assert params.ndim == 3
            assert params.shape[0] == n_samples
        assert isinstance(min_μm, (int, float))
        assert isinstance(max_μm, (int, float))
        assert isinstance(n_classes, int)
        assert isinstance(precision, int)
        assert isinstance(noise, int)
        assert min_μm > 0
        assert max_μm > 0
        assert max_μm > min_μm
        assert n_classes > 1
        assert precision > 1
        assert noise > 1
        # prepare data
        self.__target = target
        self.__distribution_type = distribution_type
        self.__params = params
        self.__n_samples, self.__n_params, self.__n_components = params.shape
        self.__n_classes = n_classes
        self.__min_μm, self.__max_μm = min_μm, max_μm
        self.__min_φ, self.__max_φ = convert_μm_to_φ(min_μm), convert_μm_to_φ(max_μm)
        self.__classes_φ = np.linspace(self.__min_φ, self.__max_φ, n_classes)
        self.__interval_φ = abs((self.__max_φ - self.__min_φ) / (n_classes - 1))
        self.__classes_μm = convert_φ_to_μm(self.__classes_φ)
        self.__precision = precision
        self.__noise = noise

        classes = np.expand_dims(np.expand_dims(self.classes_φ, 0), 0).repeat(self.n_samples, 0).repeat(self.n_components, 1)
        distribution_class = DISTRIBUTION_CLASS_MAP[distribution_type]
        proportions, components, (m, v, s, k) = distribution_class.interpret(params, classes, self.interval_φ)
        noise = np.random.randn(self.n_samples, self.n_classes) * (10**(-self.noise))
        distributions = np.round((proportions @ components).squeeze(1) + noise, decimals=self.precision)
        self.__proportions = proportions
        self.__components = components
        self.__distributions = distributions
        self.__samples = []
        for i in range(self.n_samples):
            sample = ArtificialSample(
                f"AS{i+1}",
                self.classes_μm,
                self.classes_φ,
                distributions[i],
                components[i],
                proportions[i, 0],
                (m[i], v[i], s[i], k[i]))
            self.__samples.append(sample)

    @property
    def target(self) -> typing.Iterable[typing.Iterable[typing.Tuple[float, float]]]:
        return self.__target

    @property
    def distribution_type(self) -> DistributionType:
        return self.__distribution_type

    @property
    def n_samples(self) -> int:
        return self.__n_samples

    @property
    def n_components(self) -> int:
        return self.__n_components

    @property
    def n_classes(self) -> int:
        return self.__n_classes

    @property
    def min_μm(self) -> float:
        return self.__min_μm

    @property
    def max_μm(self) -> float:
        return self.__max_μm

    @property
    def min_φ(self) -> float:
        return self.__min_φ

    @property
    def max_φ(self) -> float:
        return self.__max_φ

    @property
    def classes_φ(self) -> np.ndarray:
        return self.__classes_φ

    @property
    def classes_μm(self) -> np.ndarray:
        return self.__classes_μm

    @property
    def interval_φ(self) -> float:
        return self.__interval_φ

    @property
    def noise(self) -> int:
        return self.__noise

    @property
    def precision(self) -> int:
        return self.__precision

    @property
    def params(self) -> np.ndarray:
        return self.__params

    @property
    def samples(self) -> typing.List[ArtificialSample]:
        return self.__samples

    @property
    def distributions(self) -> np.ndarray:
        return self.__distributions

    @property
    def dataset_to_fit(self) -> GrainSizeDataset:
        dataset = GrainSizeDataset()
        sample_names = [f"AS{i+1}" for i in range(self.n_samples)]
        dataset.add_batch(self.classes_μm, sample_names, self.distributions)
        return dataset

    @property
    def proportions(self) -> np.ndarray:
        return self.__proportions

    @property
    def components(self) -> np.ndarray:
        return self.__components


def get_params(
    target: typing.Iterable[typing.Iterable[typing.Tuple[float, float]]],
    n_samples: int):
    n_components = len(target)
    n_params = len(target[0])
    params = np.random.randn(n_samples, n_params, n_components)
    for component_i, sub_target in enumerate(target):
        for param_i, (mean, std) in enumerate(sub_target):
            params[:, param_i, component_i] = params[:, param_i, component_i] * std + mean
    return params


def get_dataset(
        target: typing.Iterable[typing.Iterable[typing.Tuple[float, float]]],
        distribution_type: DistributionType,
        n_samples: int,
        min_μm=0.02,
        max_μm=2000.0,
        n_classes=101,
        precision=4,
        noise=5):
    dataset = ArtificialDataset(
        target, distribution_type,
        n_samples, params=None,
        min_μm=min_μm, max_μm=max_μm, n_classes=n_classes,
        precision=precision, noise=noise)
    return dataset


def get_sample(
        target: typing.Iterable[typing.Iterable[typing.Tuple[float, float]]],
        distribution_type: DistributionType,
        min_μm=0.02,
        max_μm=2000.0,
        n_classes=101,
        precision=4,
        noise=5):
    dataset = ArtificialDataset(
        target, distribution_type,
        1, params=None,
        min_μm=min_μm, max_μm=max_μm, n_classes=n_classes,
        precision=precision, noise=noise)
    sample=dataset.samples[0]
    return sample


def get_mean_sample(
        target: typing.Iterable[typing.Iterable[typing.Tuple[float, float]]],
        distribution_type: DistributionType,
        min_μm=0.02,
        max_μm=2000.0,
        n_classes=101,
        precision=4,
        noise=5):
    params = np.expand_dims(np.array([[mean for (mean, std) in comp] for comp in target]).T, 0)
    dataset = ArtificialDataset(
        target, distribution_type,
        1, params=params,
        min_μm=min_μm, max_μm=max_μm, n_classes=n_classes,
        precision=precision, noise=noise)
    sample=dataset.samples[0]
    return sample



class UDMAlgorithmSetting:
    def __init__(self,
                 device="cpu",
                 pretrain_epochs=400,
                 min_epochs=200,
                 max_epochs=2000,
                 precision=4.0,
                 learning_rate=1e-2,
                 betas=(0.8, 0.5),
                 constraint_level=2.0):
        assert isinstance(device, str)
        assert isinstance(pretrain_epochs, int)
        assert isinstance(min_epochs, int)
        assert isinstance(max_epochs, int)
        assert isinstance(precision, (int, float))
        assert isinstance(learning_rate, (int, float))
        assert isinstance(betas, tuple)
        assert len(betas) == 2
        beta1, beta2 = betas
        assert isinstance(beta1, float)
        assert isinstance(beta2, float)
        assert isinstance(constraint_level, (int, float))
        assert pretrain_epochs >= 0
        assert min_epochs > 0
        assert max_epochs > 0
        assert 1.0 < precision < 100.0
        assert learning_rate > 0.0
        assert 0.0 < beta1 < 1.0
        assert 0.0 < beta2 < 1.0
        assert -20 < constraint_level < 20.0
        self.device = device
        self.pretrain_epochs = pretrain_epochs
        self.min_epochs = min_epochs
        self.max_epochs = max_epochs
        self.precision = precision
        self.learning_rate = learning_rate
        self.betas = betas
        self.constraint_level = constraint_level

    def __str__(self) -> str:
        return self.__dict__.__str__()



import os

from PySide6 import QtCore, QtWidgets, QtGui

about_md = """

# QGrain

QGrain aims to provide an easy-to-use and comprehensive analysis platform for grain size distributions. QGrain has implemented many functions, however, there still are many useful tools that have not been contained. Hence, we published QGrain as an open-source project, and welcome other researchers to contribute their ideas and codes. Codes are available at this GitHub [repository](https://github.com/yuriok/QGrain/). More information and tutorials are available at the [offical website](https://qgrain.net/).

Please cite:

* Liu, Y., Liu, X., Sun, Y., 2021. QGrain: An open-source and easy-to-use software for the comprehensive analysis of grain size distributions. Sedimentary Geology 423, 105980. https://doi.org/10.1016/j.sedgeo.2021.105980

Feel free to contact the author below.

* Yuming Liu, a PhD student of IEECAS, [liuyuming@ieecas.cn](mailto:liuyuming@ieecas.cn)

"""

class AboutDialog(QtWidgets.QDialog):
    def __init__(self, parent=None):
        super().__init__(parent=parent, f=QtCore.Qt.Window)
        self.init_ui()

    def init_ui(self):
        self.setWindowTitle(self.tr("About"))
        self.setMinimumSize(400, 400)
        self.main_layout = QtWidgets.QGridLayout(self)
        self.text = QtWidgets.QTextBrowser()
        self.text.setReadOnly(True)
        self.text.setMarkdown(self.tr(about_md))
        self.text.setOpenExternalLinks(True)
        self.main_layout.addWidget(self.text, 0, 0)


    def changeEvent(self, event: QtCore.QEvent):
        if event.type() == QtCore.QEvent.LanguageChange:
            self.retranslate()

    def retranslate(self):
        self.setWindowTitle(self.tr("About"))



import logging
import string
import typing
from collections import Counter

import numpy as np
import openpyxl
from openpyxl.worksheet.worksheet import Worksheet
from sklearn.cluster import KMeans
from sklearn.decomposition import PCA

from .. import QGRAIN_VERSION
from ..artificial._generator import ArtificialDataset
from ..emma import EMMAResult
from ..model import GrainSizeDataset, GrainSizeSample
from ..ssu import DISTRIBUTION_CLASS_MAP, SSUResult, DistributionType
from ..statistic._GRADISTAT import _get_all_scales, get_all_statistic, logarithmic
from ..udm import UDMAlgorithmSetting, UDMResult
from ._use_excel import column_to_char, prepare_styles

SMALL_WIDTH = 12
MEDIAN_WIDTH = 24
LARGE_WIDTH = 48


def _check_dataset(dataset: GrainSizeDataset):
    if dataset is None:
        raise ValueError("The dataset is `None`.")
    elif not dataset.has_sample:
        raise ValueError("There is no sample in this dataset.")


def _write_readme_sheet(ws: Worksheet, text: str):
    ws.title = "README"
    full_text = \
        """
        This Excel file was generated by QGrain ({0}).

        Please cite:
        Liu, Y., Liu, X., Sun, Y., 2021. QGrain: An open-source and easy-to-use software for the comprehensive analysis of grain size distributions. Sedimentary Geology 423, 105980.
        DOI: 10.1016/j.sedgeo.2021.105980

        {1}

        """.format(QGRAIN_VERSION, text)
    lines_of_description = full_text.split("\n")
    for row, line in enumerate(lines_of_description):
        cell = ws.cell(row+1, 1, value=line)
        cell.style = "description"
    ws.column_dimensions[column_to_char(0)].width = 200


def _write_dataset_sheet(
        ws: Worksheet,
        dataset: GrainSizeDataset,
        progress_callback: typing.Callable = None):
    def write(row, col, value, style="normal_light"):
        cell = ws.cell(row+1, col+1, value=value)
        cell.style = style

    ws.title = "GSDs"
    write(0, 0, "Sample Name", style="header")
    ws.column_dimensions[column_to_char(0)].width = 24
    for col, value in enumerate(dataset.classes_μm, 1):
        write(0, col, value, style="header")
        ws.column_dimensions[column_to_char(col)].width = 10

    for i_sample, sample in enumerate(dataset.samples):
        row = i_sample + 1
        if row % 2 == 0:
            style = "normal_dark"
        else:
            style = "normal_light"
        write(row, 0, sample.name, style=style)
        for col, value in enumerate(sample.distribution, 1):
            write(row, col, value, style=style)
        if progress_callback is not None:
            progress = i_sample / dataset.n_samples
            progress_callback(progress)


def save_artificial_dataset(
        dataset: ArtificialDataset,
        filename: str,
        progress_callback: typing.Callable = None,
        logger: logging.Logger = None):
    if logger is None:
        logger = logging.getLogger("QGrain")
    else:
        assert isinstance(logger, logging.Logger)
    logger.debug("Start to save artificial dataset.")
    wb = openpyxl.Workbook()
    prepare_styles(wb)
    logger.debug("Creating the `README` sheet.")
    readme_text = \
        """
        It contanins [number of components + 3] sheets:
        1. The first sheet is used to put the random settings which were used to generate random parameters.
        2. The second sheet is the generated dataset.
        3. The third sheet stores the random parameters which were used to calulate the component distributions and their mixture.
        4. The left sheets are the distributions of all components.

        Sampling Settings
            Minimum size [μm]: {0}
            Maximum size [μm]: {1}
            Number of Grain Size Classes: {2}
            Precision: {3}
            Noise Decimals: {4}
            Number of Samples: {5}
        """.format(
            dataset.min_μm,
            dataset.max_μm,
            dataset.n_classes,
            dataset.precision,
            dataset.noise,
            dataset.n_samples)
    _write_readme_sheet(wb.active, readme_text)

    def write(row, col, value, style="normal_light"):
        cell = ws.cell(row+1, col+1, value=value)
        cell.style = style

    param_names = list(DISTRIBUTION_CLASS_MAP[dataset.distribution_type].PARAM_NAMES) + ["Weight"]
    n_params = DISTRIBUTION_CLASS_MAP[dataset.distribution_type].N_PARAMS + 1
    if dataset.target is not None:
        logger.debug("Creating the `Random Settings` sheet.")
        ws = wb.create_sheet("Random Settings")
        write(0, 0, "Parameter", style="header")
        ws.merge_cells(start_row=1, start_column=1, end_row=2, end_column=1)
        for i_param, param_name in enumerate(param_names):
            write(0, i_param*2+1, param_name, style="header")
            ws.merge_cells(start_row=1, start_column=i_param*2+2, end_row=1, end_column=i_param*2+3)
        ws.column_dimensions[column_to_char(0)].width = 16
        for col in range(1, n_params*2+1):
            ws.column_dimensions[column_to_char(col)].width = 16
            if col % 2 == 0:
                write(1, col, "Standard deviation", style="header")
            else:
                write(1, col, "Mean", style="header")
        for row, comp_params in enumerate(dataset.target, 2):
            if row % 2 == 1:
                style = "normal_dark"
            else:
                style = "normal_light"
            write(row, 0, f"Component{row-1}", style=style)
            for i, (mean, std) in enumerate(comp_params):
                write(row, i*2+1, mean, style=style)
                write(row, i*2+2, std, style=style)
    else:
        logger.warning("Not creating the `Random Settings` sheet.")

    logger.debug("Creating the `Dataset` sheet.")
    ws = wb.create_sheet("Dataset")
    write(0, 0, "Sample Name", style="header")
    ws.column_dimensions[column_to_char(0)].width = 24
    for col, value in enumerate(dataset.classes_μm, 1):
        write(0, col, value, style="header")
        ws.column_dimensions[column_to_char(col)].width = 10
    for i, sample in enumerate(dataset.samples):
        row = i + 1
        if row % 2 == 0:
            style = "normal_dark"
        else:
            style = "normal_light"
        write(row, 0, sample.name, style=style)
        for col, value in enumerate(sample.distribution, 1):
            write(row, col, value, style=style)

        if progress_callback is not None:
            progress_callback(i / dataset.n_samples * 0.2)

    logger.debug("Creating the `Parameters` sheet.")
    ws = wb.create_sheet("Parameters")
    write(0, 0, "Sample Name", style="header")
    ws.merge_cells(start_row=1, start_column=1, end_row=2, end_column=1)
    ws.column_dimensions[column_to_char(0)].width = 24
    for i in range(dataset.n_components):
        write(0, n_params*i+1, f"Component{i+1}", style="header")
        ws.merge_cells(start_row=1, start_column=n_params*i+2, end_row=1, end_column=n_params*(i+1)+1)
        for j, header_name in enumerate(param_names):
            write(1, n_params*i+1+j, header_name, style="header")
            ws.column_dimensions[column_to_char(n_params*i+1+j)].width = 16
    for i in range(dataset.n_samples):
        row = i + 2
        if row % 2 == 1:
            style = "normal_dark"
        else:
            style = "normal_light"
        write(row, 0, dataset.samples[i].name, style=style)
        for j in range(dataset.n_components):
            for k in range(n_params):
                write(row, n_params*j+k+1, dataset.params[i, k, j], style=style)

        if progress_callback is not None:
            progress_callback(1/dataset.n_samples*0.2 + 0.2)

    for i in range(dataset.n_components):
        logger.debug(f"Creating the `C{i+1}` sheet.")
        ws = wb.create_sheet(f"C{i+1}")
        write(0, 0, "Sample Name", style="header")
        ws.column_dimensions[column_to_char(0)].width = 24
        for col, value in enumerate(dataset.classes_μm, 1):
            write(0, col, value, style="header")
            ws.column_dimensions[column_to_char(col)].width = 10
        for row, sample in enumerate(dataset.samples, 1):
            if row % 2 == 0:
                style = "normal_dark"
            else:
                style = "normal_light"
            write(row, 0, sample.name, style=style)
            for col, value in enumerate(sample.components[i].distribution, 1):
                write(row, col, value, style=style)

        if progress_callback is not None:
            progress_callback(((i*dataset.n_samples + row) / dataset.n_samples*dataset.n_components) * 0.6 + 0.4)
    wb.save(filename)
    wb.close()
    if progress_callback is not None:
        progress_callback(1.0)
    logger.info(f"The artificial dataset has been saved to the Excel file: [{filename}].")


def save_dataset(
        dataset: GrainSizeDataset,
        filename: str,
        progress_callback: typing.Callable = None,
        logger: logging.Logger = None):
    _check_dataset(dataset)
    if logger is None:
        logger = logging.getLogger("QGrain")
    else:
        assert isinstance(logger, logging.Logger)
    logger.debug("Start to save grain size dataset.")
    wb = openpyxl.Workbook()
    prepare_styles(wb)
    logger.debug("Creating the `README` sheet.")
    readme_text = \
        """
        It only contanins one sheet which stores the grain size distributions.
        """
    _write_readme_sheet(wb.active, readme_text)
    logger.debug("Creating the `GSDs` sheet.")
    ws = wb.create_sheet("GSDs")
    _write_dataset_sheet(ws, dataset, progress_callback)
    logger.debug("Saving the workbook to file.")
    wb.save(filename)
    wb.close()
    if progress_callback is not None:
        progress_callback(1.0)
    logger.info(f"The dataset has been saved to the Excel file: [{filename}].")


def save_statistic(
        dataset: GrainSizeDataset,
        filename: str,
        progress_callback: typing.Callable = None,
        logger: logging.Logger = None):
    _check_dataset(dataset)
    if logger is None:
        logger = logging.getLogger("QGrain")
    else:
        assert isinstance(logger, logging.Logger)
    logger.debug("Start to save statistic result.")
    # Calculate
    logger.debug("Calculating the statistic parameters and classification groups of all samples.")
    all_statistics = []
    for i, sample in enumerate(dataset.samples):
        sample_statistics = get_all_statistic(sample.classes_μm, sample.classes_φ, sample.distribution)
        all_statistics.append(sample_statistics)
        if progress_callback is not None:
            progress = (i / dataset.n_samples) * 0.4
            progress_callback(progress)
    wb = openpyxl.Workbook()
    prepare_styles(wb)
    logger.debug("Creating the `README` sheet.")
    readme_text = \
        """
        It contanins 6 sheets.
            1-5. The previous five sheets stores the statistic parameters of five computational methods, respectively.
            6. The last sheet puts the proportions of different size scales and the classification groups.

        The statistic formulas are referred to Blott & Pye (2001)'s work.
        The classification of GSDs is referred to Folk (1954)'s and Blott & Pye (2012)'s scheme.

        References:
            1. Blott, S. J. & Pye, K. Particle size scales and classification of sediment types based on particle size distributions: Review and recommended procedures. Sedimentology 59, 2071–2096 (2012).
            2. Blott, S. J. & Pye, K. GRADISTAT: a grain-size distribution and statistics package for the analysis of unconsolidated sediments. Earth Surf. Process. Landforms 26, 1237–1248 (2001).
            3. Folk, R. L. The Distinction between Grain Size and Mineral Composition in Sedimentary-Rock Nomenclature. The Journal of Geology 62, 344–359 (1954).
        """
    _write_readme_sheet(wb.active, readme_text)
    logger.debug("Creating the `GSDs` sheet.")
    if progress_callback is not None:
        _callback = lambda progress: progress_callback(progress * 0.1 + 0.4)
    else:
        _callback = None
    ws = wb.create_sheet("GSDs")
    _write_dataset_sheet(ws, dataset, progress_callback=_callback)

    def get_keys(method: str):
        small_width = 12
        median_width = 24
        large_width = 48
        if method == "arithmetic":
            unit = "μm"
            keys = [
                (lambda s: s[method]["mean"], f"Mean [{unit}]", small_width),
                (lambda s: s[method]["std"], f"Sorting Coefficient", small_width),
                (lambda s: s[method]["skewness"], f"Skewness", small_width),
                (lambda s: s[method]["kurtosis"], f"Kurtosis", small_width)]
            return keys
        elif method in ("geometric", "logarithmic", "geometric_FW57", "logarithmic_FW57"):
            unit = "μm" if method.startswith("geometric") else "φ"
            keys = [
                (lambda s: s[method]["mean"], f"Mean [{unit}]", small_width),
                (lambda s: s[method]["mean_description"], "Mean Description", median_width),
                (lambda s: s[method]["median"], f"Median [{unit}]", small_width),
                (lambda s: s[method]["mode"], f"Mode [{unit}]", small_width),
                (lambda s: len(s[method]["modes"]), f"Number of Modes", small_width),
                (lambda s: ", ".join([f"{mode: 0.4f}" for mode in s[method]["modes"]]), f"Modes [{unit}]", median_width),
                (lambda s: s[method]["std"], "Sorting Coefficient", small_width),
                (lambda s: s[method]["std_description"], "Sorting Description", median_width),
                (lambda s: s[method]["skewness"], "Skewness", small_width),
                (lambda s: s[method]["skewness_description"], "Skewness Description", median_width),
                (lambda s: s[method]["kurtosis"], "Kurtosis", small_width),
                (lambda s: s[method]["kurtosis_description"], "Kurtosis Description", median_width)]
            return keys
        elif method == "proportion_and_classification":
            keys = [
                (lambda s: ", ".join([f"{p*100:0.4f}" for p in s["GSM_proportion"]]), "(Gravel, Sand, Mud) Proportions [%]", large_width),
                (lambda s: ", ".join([f"{p*100:0.4f}" for p in s["SSC_proportion"]]), "(Sand, Silt, Clay) Proportions [%]", large_width),
                (lambda s: ", ".join([f"{p*100:0.4f}" for p in s["BGSSC_proportion"]]), "(Boulder, Gravel, Sand, Silt, Clay) Proportions [%]", large_width),
                (lambda s: s["group_Folk54"], "Group (Folk, 1954)", median_width),
                (lambda s: s["group_BP12"], "Group (Blott & Pye, 2012)", large_width),
                (lambda s: s["group_BP12_symbol"], "Group Symbol (Blott & Pye, 2012)", median_width)]
            all_scales = _get_all_scales()
            for scale in all_scales:
                func = lambda s, scale=scale: s["proportion"][scale] * 100.0
                name = string.capwords(" ".join(scale)) + " Proportion [%]"
                keys.append((func, name, small_width))
            return keys
        else:
            raise NotImplementedError(method)

    def write(row, col, value, style="normal_light"):
        cell = ws.cell(row+1, col+1, value=value)
        cell.style = style

    methods = ["arithmetic", "geometric", "logarithmic", "geometric_FW57", "logarithmic_FW57", "proportion_and_classification"]
    sheet_names = ["Arithmetic", "Geometric", "Logarithmic", "Geometric_FW57", "Logarithmic_FW57", "Proportion and Classification"]
    for i_method, (method, sheet_name) in enumerate(zip(methods, sheet_names)):
        logger.debug(f"Creating the `{sheet_name}` sheet.")
        ws = wb.create_sheet(sheet_name)
        write(0, 0, "Sample Name", style="header")
        keys = get_keys(method)
        ws.column_dimensions[column_to_char(0)].width = 16
        for col, (func, name, width) in enumerate(keys, 1):
            write(0, col, name, style="header")
            ws.column_dimensions[column_to_char(col)].width = width
        for i_sample, (sample, sample_statistics) in enumerate(zip(dataset.samples, all_statistics)):
            row = i_sample + 1
            if row % 2 == 0:
                style = "normal_dark"
            else:
                style = "normal_light"
            write(row, 0, sample.name, style=style)
            for col, (func, name, width) in enumerate(keys, 1):
                value = func(sample_statistics)
                write(row, col, value, style=style)
            if progress_callback is not None:
                progress = 0.5 + ((i_sample / dataset.n_samples) + i_method) / len(methods) * 0.5
                progress_callback(progress)
    logger.debug("Saving the workbook to file.")
    wb.save(filename)
    wb.close()
    if progress_callback is not None:
        progress_callback(1.0)
    logger.info(f"The statistic result has been saved to the Excel file: [{filename}].")


def save_pca(
        dataset: GrainSizeDataset,
        filename: str,
        progress_callback: typing.Callable = None,
        logger: logging.Logger = None):
    _check_dataset(dataset)
    if logger is None:
        logger = logging.getLogger("QGrain")
    else:
        assert isinstance(logger, logging.Logger)
    logger.debug("Start to save PCA result.")
    # Calculate
    logger.debug("Performing the PCA algorithm to this dataset. The number of PCs is set to 10.")
    pca = PCA(n_components=10)
    transformed = pca.fit_transform(dataset.distribution_matrix)
    components = pca.components_
    ratios = pca.explained_variance_ratio_

    wb = openpyxl.Workbook()
    prepare_styles(wb)
    logger.debug("Creating the `README` sheet.")
    readme_text = \
        """
        It contanins 3 sheets:
        1. The first sheet is the dataset which was used to perform the PCA algorithm.
        2. The second sheet is used to put the distributions of all PCs.
        3. The third sheet is used to store the PC variations of all samples.

        The base PCA algorithm is implemented by scikit-learn. You can get the details of algorithm from the following website.
        https://scikit-learn.org/stable/modules/generated/sklearn.decomposition.PCA.html
        """
    _write_readme_sheet(wb.active, readme_text)
    logger.debug("Creating the `GSDs` sheet.")
    if progress_callback is not None:
        _callback = lambda progress: progress_callback(progress * 0.2)
    else:
        _callback = None
    ws = wb.create_sheet("GSDs")
    _write_dataset_sheet(ws, dataset, progress_callback=_callback)

    def write(row, col, value, style="normal_light"):
        cell = ws.cell(row+1, col+1, value=value)
        cell.style = style

    logger.debug("Creating the `Distributions of PCs` sheet.")
    ws = wb.create_sheet("Distributions of PCs")
    write(0, 0, "PC", style="header")
    ws.column_dimensions[column_to_char(0)].width = 16
    for col, value in enumerate(dataset.classes_μm, 1):
        write(0, col, value, style="header")
        ws.column_dimensions[column_to_char(col)].width = 10
    for i, component in enumerate(components):
        row = i + 1
        if row % 2 == 0:
            style = "normal_dark"
        else:
            style = "normal_light"
        write(row, 0, f"PC{i+1} ({ratios[i]:0.2%})", style=style)
        for col, value in enumerate(component, 1):
            write(row, col, value, style=style)
    if progress_callback is not None:
        progress_callback(0.3)

    logger.debug("Creating the `Variations of PCs` sheet.")
    ws = wb.create_sheet("Variations of PCs")
    write(0, 0, "Sample Name", style="header")
    ws.column_dimensions[column_to_char(0)].width = 16
    for i in range(10):
        write(0, i+1, f"PC{i+1} ({ratios[i]:0.2%})", style="header")
        ws.column_dimensions[column_to_char(i+1)].width = 10
    for row, varations in enumerate(transformed, 1):
        if row % 2 == 0:
            style = "normal_dark"
        else:
            style = "normal_light"
        write(row, 0, dataset.samples[row-1].name, style=style)
        for col, value in enumerate(varations, 1):
            write(row, col, value, style=style)
        if progress_callback is not None:
            progress_callback(row / dataset.n_samples * 0.7 + 0.3)

    wb.save(filename)
    wb.close()
    if progress_callback is not None:
        progress_callback(1.0)
    logger.info(f"The PCA result has been saved to the Excel file: [{filename}].")


def save_clustering(
        dataset: GrainSizeDataset,
        flags: typing.Iterable[int],
        filename: str,
        progress_callback: typing.Callable = None,
        logger: logging.Logger = None):
    _check_dataset(dataset)
    if logger is None:
        logger = logging.getLogger("QGrain")
    else:
        assert isinstance(logger, logging.Logger)
    logger.debug("Start to save clustering result.")
    flag_set = set(flags)
    n_clusters = len(flag_set)
    typical_samples = [] # type: list[GrainSizeSample]
    temp_flag_set = set()
    for i, flag in enumerate(flags):
        if len(temp_flag_set) == n_clusters:
            break
        if flag not in temp_flag_set:
            typical_samples.append(dataset.samples[i])
            temp_flag_set.add(flag)

    wb = openpyxl.Workbook()
    prepare_styles(wb)
    logger.debug("Creating the `README` sheet.")
    readme_text = \
        """
        It contanins 3 or [number of clusters + 3] sheets:
        1. The first sheet is the dataset which was used to perform the hierarchy clustering algorithm.
        2. The second sheet is used to put the clustering flags of all samples.
        3. The third sheet is the typical sampels (i.e, the first sample of each cluster was selected).
        4. If the number of clusters less equal to 100, the samples of each cluster will be save to individual sheets.

        The base hierarchy clusrting algorithm is implemented by Scipy. You can get the details of algorithm from the following website.
        https://docs.scipy.org/doc/scipy/reference/cluster.hierarchy.html
        """
    _write_readme_sheet(wb.active, readme_text)
    logger.debug("Creating the `GSDs` sheet.")
    if progress_callback is not None:
        _callback = lambda progress: progress_callback(progress * 0.2)
    else:
        _callback = None
    ws = wb.create_sheet("GSDs")
    _write_dataset_sheet(ws, dataset, progress_callback=_callback)

    def write(row, col, value, style="normal_light"):
        cell = ws.cell(row+1, col+1, value=value)
        cell.style = style

    logger.debug("Creating the `Cluster Flags of Samples` sheet.")
    ws = wb.create_sheet("Cluster Flags of Samples")
    write(0, 0, "Sample Name", style="header")
    write(0, 1, "Cluster Flags", style="header")
    ws.column_dimensions[column_to_char(0)].width = 16
    ws.column_dimensions[column_to_char(1)].width = 16
    for i, (sample, flag) in enumerate(zip(dataset.samples, flags)):
        row = i + 1
        if row % 2 == 0:
            style = "normal_dark"
        else:
            style = "normal_light"
        write(row, 0, sample.name, style=style)
        write(row, 1, flag, style=style)
        if progress_callback is not None:
            if n_clusters <= 100:
                progress_callback(i / dataset.n_samples * 0.1 + 0.2)
            else:
                progress_callback(i / dataset.n_samples * 0.4 + 0.2)

    logger.debug("Creating the `Typical Samples of Clusters` sheet.")
    ws = wb.create_sheet("Typical Samples of Clusters")
    write(0, 0, "Sample Name", style="header")
    ws.column_dimensions[column_to_char(0)].width = 16
    for col, value in enumerate(dataset.classes_μm, 1):
        write(0, col, value, style="header")
        ws.column_dimensions[column_to_char(col)].width = 10
    for i, sample in enumerate(typical_samples):
        row = i + 1
        if row % 2 == 0:
            style = "normal_dark"
        else:
            style = "normal_light"
        write(row, 0, sample.name, style=style)
        for col, value in enumerate(sample.distribution, 1):
            write(row, col, value, style=style)
        if progress_callback is not None:
            if n_clusters <= 100:
                progress_callback(i / dataset.n_samples * 0.1 + 0.3)
            else:
                progress_callback(i / dataset.n_samples * 0.4 + 0.6)

    if n_clusters <= 100:
        for flag in flag_set:
            samples = []
            for sample, in_this_cluster in zip(dataset.samples, np.equal(flags, flag)):
                if in_this_cluster:
                    samples.append(sample)
            logger.debug(f"Creating the `Cluster{flag}` sheet.")
            ws = wb.create_sheet(f"Cluster{flag}")
            write(0, 0, "Sample Name", style="header")
            ws.column_dimensions[column_to_char(0)].width = 16
            for col, value in enumerate(dataset.classes_μm, 1):
                write(0, col, value, style="header")
                ws.column_dimensions[column_to_char(col)].width = 10
            for i, sample in enumerate(samples):
                row = i + 1
                if row % 2 == 0:
                    style = "normal_dark"
                else:
                    style = "normal_light"
                write(row, 0, sample.name, style=style)
                for col, value in enumerate(sample.distribution, 1):
                    write(row, col, value, style=style)
                if progress_callback is not None:
                    progress_callback(i / dataset.n_samples / n_clusters * 0.6 + 0.4)

    wb.save(filename)
    wb.close()
    if progress_callback is not None:
        progress_callback(1.0)
    logger.info(f"The Clustering result has been saved to the Excel file: [{filename}].")


def save_emma(
        result: EMMAResult, filename: str,
        progress_callback: typing.Callable = None,
        logger: logging.Logger = None):
    if logger is None:
        logger = logging.getLogger("QGrain")
    else:
        assert isinstance(logger, logging.Logger)

    logger.debug("Start to save EMMA result.")
    # get the mode size of each end-members
    modes = [(i, result.dataset.classes_μm[np.unravel_index(np.argmax(result.end_members[i]), result.end_members[i].shape)]) for i in range(result.n_members)]
    # sort them by mode size
    modes.sort(key=lambda x: x[1])

    wb = openpyxl.Workbook()
    prepare_styles(wb)
    logger.debug("Creating the `README` sheet.")
    readme_text = \
        """
        It contanins three sheets:
        1. The first sheet is the dataset which was used to perform the EMMA algorithm.
        2. The second sheet is used to put the distributions of all end members.
        3. The third sheet is the proportions of end members of all samples.

        This EMMA algorithm was implemented by QGrian, using the famous machine learning framework, PyTorch.

        EMMA Algorithm Details
            Number of Samples: {0}
            Kernel Type: {1}
            Number of End Members: {2}
            Initial Parameters: {3}
            Number of Epochs: {4}
            Spent Time: {5} s

            Computing Device: {6}
            Distance Function: {7}
            Number of Pretrain Epochs: {8}
            Minimum Number of Epochs: {9}
            Maximum Number of Epochs: {10}
            Learning Rate: {11}
            Precision: {12}
            Betas: {13}

        """.format(
            result.n_samples,
            result.kernel_type.name,
            result.n_members,
            result.initial_parameters,
            result.n_iterations,
            result.time_spent,
            result.resolver_setting.device,
            result.resolver_setting.distance,
            result.resolver_setting.pretrain_epochs,
            result.resolver_setting.min_epochs,
            result.resolver_setting.max_epochs,
            result.resolver_setting.learning_rate,
            result.resolver_setting.precision,
            result.resolver_setting.betas)
    _write_readme_sheet(wb.active, readme_text)
    logger.debug("Creating the `GSDs` sheet.")
    if progress_callback is not None:
        _callback = lambda progress: progress_callback(progress * 0.4)
    else:
        _callback = None
    ws = wb.create_sheet("GSDs")
    _write_dataset_sheet(ws, result.dataset, progress_callback=_callback)

    def write(row, col, value, style="normal_light"):
        cell = ws.cell(row+1, col+1, value=value)
        cell.style = style

    logger.debug("Creating the `Distributions of End Members` sheet.")
    ws = wb.create_sheet("Distributions of End Members")
    write(0, 0, "End Member", style="header")
    ws.column_dimensions[column_to_char(0)].width = 16
    for col, value in enumerate(result.dataset.classes_μm, 1):
        write(0, col, value, style="header")
        ws.column_dimensions[column_to_char(col)].width = 10
    for i, (index, _) in enumerate(modes):
        row = i + 1
        if row % 2 == 0:
            style = "normal_dark"
        else:
            style = "normal_light"
        write(row, 0, f"EM{i+1}", style=style)
        for col, value in enumerate(result.end_members[index], 1):
            write(row, col, value, style=style)

    logger.debug("Creating the `Proportions of End Members` sheet.")
    ws = wb.create_sheet("Proportions of End Members")
    write(0, 0, "Sample Name", style="header")
    ws.column_dimensions[column_to_char(0)].width = 16
    for i in range(result.n_members):
        write(0, i+1, f"EM{i+1}", style="header")
        ws.column_dimensions[column_to_char(i+1)].width = 10
    for i, sample_proportions in enumerate(result.proportions):
        row = i + 1
        if row % 2 == 0:
            style = "normal_dark"
        else:
            style = "normal_light"
        write(row, 0, result.dataset.samples[i].name, style=style)
        for col, (index, _) in enumerate(modes, 1):
            write(row, col, sample_proportions[index], style=style)
        if progress_callback is not None:
            progress_callback(i / result.n_samples * 0.6 + 0.4)

    wb.save(filename)
    wb.close()
    if progress_callback is not None:
        progress_callback(1.0)
    logger.info(f"The EMMA result has been saved to the Excel file: [{filename}].")


def save_ssu(
        results: typing.List[SSUResult], filename: str,
        align_components=False,
        progress_callback: typing.Callable = None,
        logger: logging.Logger = None):
    if logger is None:
        logger = logging.getLogger("QGrain")
    else:
        assert isinstance(logger, logging.Logger)
    logger.debug("Start to save SSU results.")
    # pack the GSDs of samples to the dataset
    dataset = GrainSizeDataset()
    dataset.add_batch(
        results[0].classes_μm,
        [result.sample.name for result in results],
        [result.sample.distribution for result in results],
        need_validation=False)
    max_n_components = max(Counter([result.n_components for result in results]).keys())

    # prepare flags
    flags = []
    if not align_components:
        for result in results:
            flags.extend(range(result.n_components))
    else:
        stacked_components = []
        for result in results:
            for component in result.components:
                stacked_components.append(component.distribution)
        stacked_components = np.array(stacked_components)
        clusers = KMeans(n_clusters=max_n_components)
        flags = clusers.fit_predict(stacked_components)
        # check flags to make it unique
        flag_index = 0
        for result in results:
            result_flags = set()
            for component in result.components:
                if flags[flag_index] in result_flags:
                    if flags[flag_index] == max_n_components:
                        flags[flag_index] = max_n_components-1
                    else:
                        flag_index[flag_index] += 1
                    result_flags.add(flags[flag_index])
                flag_index += 1

        flag_set = set(flags)
        picked = []
        for target_flag in flag_set:
            for i, flag in enumerate(flags):
                if flag == target_flag:
                    picked.append((target_flag, logarithmic(dataset.classes_φ, stacked_components[i])["mean"]))
                    break
        picked.sort(key=lambda x: x[1])
        flag_map = {flag: index for index, (flag, _) in enumerate(picked)}
        flags = np.array([flag_map[flag] for flag in flags])

    wb = openpyxl.Workbook()
    prepare_styles(wb)
    logger.debug("Creating the `README` sheet.")
    readme_text = \
        """
        It contanins [max number of components + 4] sheets:
        1. The first sheet is used to put the grain size distributions of corresponding samples.
        2. The second sheet is used to put the fitting information and resolved parameters of SSU results.
        3. The third sheet stores the statistic parameters of all components.
        4. The fouth sheet is used to put the distributions of unmixed components and the sum.
        5. Other sheets severally store the distributions of each component group.

        The SSU algorithm is implemented by QGrain.

        """
    _write_readme_sheet(wb.active, readme_text)
    logger.debug("Creating the `GSDs` sheet.")
    if progress_callback is not None:
        _callback = lambda progress: progress_callback(progress * 0.1)
    else:
        _callback = None
    ws = wb.create_sheet("GSDs")
    _write_dataset_sheet(ws, dataset, progress_callback=_callback)

    def write(row, col, value, style="normal_light"):
        cell = ws.cell(row+1, col+1, value=value)
        cell.style = style

    logger.debug("Creating the `Information of Fitting` sheet.")
    ws = wb.create_sheet("Information of Fitting")
    write(0, 0, "Sample Name", style="header")
    ws.column_dimensions[column_to_char(0)].width = 16
    headers = ["Distribution Type",
               "Number of Components",
               "Resolver Settings",
               "Initial Guess",
               "Resolved Parameters",
               "Spent Time [s]",
               "Number of Iterations",
               "Final Distance [log10MSE]"]
    for col, value in enumerate(headers, 1):
        write(0, col, value, style="header")
        if col in (3, 4, 5):
            ws.column_dimensions[column_to_char(col)].width = 32
        else:
            ws.column_dimensions[column_to_char(col)].width = 10
    for i, result in enumerate(results):
        row = i + 1
        if row % 2 == 0:
            style = "normal_dark"
        else:
            style = "normal_light"
        write(row, 0, result.sample.name, style=style)
        write(row, 1, result.distribution_type.name, style=style)
        write(row, 2, result.n_components, style=style)
        write(row, 3, "Default" if result.task.resolver_setting is None else result.task.resolver_setting.__str__(), style=style)
        write(row, 4, "None" if result.task.initial_guess is None else result.task.initial_guess.__str__(), style=style)
        write(row, 5, result.func_args.__str__(), style=style)
        write(row, 6, result.time_spent, style=style)
        write(row, 7, result.n_iterations, style=style)
        write(row, 8, result.get_distance("log10MSE"), style=style)
        if progress_callback is not None:
            progress_callback(i / dataset.n_samples * 0.1 + 0.1)

    logger.debug("Creating the `Statistic Moments` sheet.")
    ws = wb.create_sheet("Statistic Moments")
    write(0, 0, "Sample Name", style="header")
    ws.merge_cells(start_row=1, start_column=1, end_row=2, end_column=1)
    ws.column_dimensions[column_to_char(0)].width = 16
    headers = []
    sub_headers = ["Proportion",
                   "Mean [φ]",
                   "Mean [μm]",
                   "Standard Deviation [φ]",
                   "Standard Deviation [μm]",
                   "Skewness",
                   "Kurtosis"]
    for i in range(max_n_components):
        write(0, i*len(sub_headers)+1, f"C{i+1}", style="header")
        ws.merge_cells(start_row=1, start_column=i*len(sub_headers)+2, end_row=1, end_column=(i+1)*len(sub_headers)+1)
        headers.extend(sub_headers)
    for col, value in enumerate(headers, 1):
        write(1, col, value, style="header")
        ws.column_dimensions[column_to_char(col)].width = 10
    flag_index = 0
    for i, result in enumerate(results):
        row = i + 2
        if row % 2 == 0:
            style = "normal_light"
        else:
            style = "normal_dark"
        write(row, 0, result.sample.name, style=style)
        for component in result.components:
            index = flags[flag_index]
            s = get_all_statistic(result.classes_μm, result.classes_φ, component.distribution)
            write(row, index*len(sub_headers)+1, component.proportion, style=style)
            write(row, index*len(sub_headers)+2, s["logarithmic"]["mean"], style=style)
            write(row, index*len(sub_headers)+3, s["geometric"]["mean"], style=style)
            write(row, index*len(sub_headers)+4, s["logarithmic"]["std"], style=style)
            write(row, index*len(sub_headers)+5, s["geometric"]["std"], style=style)
            write(row, index*len(sub_headers)+6, s["logarithmic"]["skewness"], style=style)
            write(row, index*len(sub_headers)+7, s["logarithmic"]["kurtosis"], style=style)
            flag_index += 1
        if progress_callback is not None:
            progress_callback(i / dataset.n_samples * 0.1 + 0.2)

    logger.debug("Creating the `Unmixed Components` sheet.")
    ws = wb.create_sheet("Unmixed Components")
    ws.merge_cells(start_row=1, start_column=1, end_row=1, end_column=2)
    write(0, 0, "Sample Name", style="header")
    ws.column_dimensions[column_to_char(0)].width = 16
    for col, value in enumerate(dataset.classes_μm, 2):
        write(0, col, value, style="header")
        ws.column_dimensions[column_to_char(col)].width = 10
    row = 1
    for i, result in enumerate(results):
        if i % 2 == 0:
            style = "normal_light"
        else:
            style = "normal_dark"
        write(row, 0, result.sample.name, style=style)
        ws.merge_cells(start_row=row+1, start_column=1, end_row=row+result.n_components+1, end_column=1)
        for component_i, component in enumerate(result.components, 1):
            write(row, 1, f"C{component_i}", style=style)
            for col, value in enumerate(component.distribution*component.proportion, 2):
                write(row, col, value, style=style)
            row += 1
        write(row, 1, "Sum", style=style)
        for col, value in enumerate(result.distribution, 2):
            write(row, col, value, style=style)
        row += 1
        if progress_callback is not None:
            progress_callback(i / dataset.n_samples * 0.2 + 0.3)

    logger.debug("Creating separate sheets for all components.")
    ws_dict = {}
    flag_set = set(flags)
    for flag in flag_set:
        ws = wb.create_sheet(f"C{flag+1}")
        write(0, 0, "Sample Name", style="header")
        ws.column_dimensions[column_to_char(0)].width = 16
        for col, value in enumerate(dataset.classes_μm, 1):
            write(0, col, value, style="header")
            ws.column_dimensions[column_to_char(col)].width = 10
        ws_dict[flag] = ws

    flag_index = 0
    for i, result in enumerate(results):
        row = i + 1
        if row % 2 == 0:
            style = "normal_dark"
        else:
            style = "normal_light"

        for component in result.components:
            flag = flags[flag_index]
            ws = ws_dict[flag]
            write(row, 0, result.sample.name, style=style)
            for col, value in enumerate(component.distribution, 1):
                write(row, col, value, style=style)
            flag_index += 1
        if progress_callback is not None:
            progress_callback(i / dataset.n_samples * 0.5 + 0.5)

    wb.save(filename)
    wb.close()
    if progress_callback is not None:
        progress_callback(1.0)
    logger.info(f"The SSU results have been saved to the Excel file: [{filename}].")


def save_udm(
        result: UDMResult, filename: str,
        progress_callback: typing.Callable = None,
        logger: logging.Logger = None):
    if logger is None:
        logger = logging.getLogger("QGrain")
    else:
        assert isinstance(logger, logging.Logger)
    logger.debug("Start to save the UDM result.")

    wb = openpyxl.Workbook()
    prepare_styles(wb)
    logger.debug("Creating the `README` sheet.")
    readme_text = \
        """
        It contanins [number of components + 4] sheets:
        1. The first sheet is used to put the grain size distributions of corresponding samples.
        2. The second sheet is used to put the resolved parameters of all samples.
        3. The third sheet stores the statistic parameters of all components.
        4. The fouth sheet is used to put the distributions of unmixed components and the sum.
        5. Other sheets severally store the distributions of each component group.

        The UDM algorithm was proposed and implemented by QGrain.

        UDM Algorithm Details
            Number of Samples: {0}
            Kernel Type: {1}
            Number of Components: {2}
            Initial Parameters: {3}
            Number of Epochs: {4}
            Spent Time: {5} s

            Computing Device: {6}
            Number of Pretrain Epochs: {7}
            Minimum Number of Epochs: {8}
            Maximum Number of Epochs: {9}
            Learning Rate: {10}
            Precision: {11}
            Betas: {12}
            Constraint Level: {13}

        """.format(
            result.n_samples,
            result.kernel_type.name,
            result.n_components,
            result.initial_parameters,
            result.n_iterations,
            result.time_spent,
            result.resolver_setting.device,
            result.resolver_setting.pretrain_epochs,
            result.resolver_setting.min_epochs,
            result.resolver_setting.max_epochs,
            result.resolver_setting.learning_rate,
            result.resolver_setting.precision,
            result.resolver_setting.betas,
            result.resolver_setting.constraint_level)
    _write_readme_sheet(wb.active, readme_text)
    logger.debug("Creating the `GSDs` sheet.")
    if progress_callback is not None:
        _callback = lambda progress: progress_callback(progress * 0.1)
    else:
        _callback = None
    ws = wb.create_sheet("GSDs")
    _write_dataset_sheet(ws, result.dataset, progress_callback=_callback)

    def write(row, col, value, style="normal_light"):
        cell = ws.cell(row+1, col+1, value=value)
        cell.style = style

    logger.debug("Creating the `Resolved Paramters` sheet.")
    ws = wb.create_sheet("Resolved Paramters")
    write(0, 0, "Sample Name", style="header")
    ws.merge_cells(start_row=1, start_column=1, end_row=2, end_column=1)
    ws.column_dimensions[column_to_char(0)].width = 16
    headers = []
    distribution_type = DistributionType._member_map_[result.kernel_type.name]
    distribution_class = DISTRIBUTION_CLASS_MAP[distribution_type]
    sub_headers = (*distribution_class.PARAM_NAMES, "Weight")
    for i in range(result.n_components):
        write(0, i*len(sub_headers)+1, f"C{i+1}", style="header")
        ws.merge_cells(start_row=1, start_column=i*len(sub_headers)+2, end_row=1, end_column=(i+1)*len(sub_headers)+1)
        headers.extend(sub_headers)
    for col, value in enumerate(headers, 1):
        write(1, col, value, style="header")
        ws.column_dimensions[column_to_char(col)].width = 10
    mean = np.mean(result.components, axis=0)
    modes = [(i, result.dataset.classes_μm[np.unravel_index(np.argmax(mean[i]), mean[i].shape)]) for i in range(result.n_components)]
    modes.sort(key=lambda x: x[1])
    sort_map = {i_c: i for i_c, (i, _) in enumerate(modes)}
    for i, sample in enumerate(result.dataset.samples):
        row = i + 2
        if row % 2 == 0:
            style = "normal_light"
        else:
            style = "normal_dark"
        write(row, 0, sample.name, style=style)
        for j in range(result.n_components):
            true_index = sort_map[j]
            for k in range(len(sub_headers)):
                col = j * len(sub_headers) + k + 1
                value = result.final_parameters[i, k, true_index]
                write(row, col, value, style=style)
        if progress_callback is not None:
            progress_callback(i / result.n_samples * 0.1 + 0.1)

    logger.debug("Creating the `Statistic Moments` sheet.")
    ws = wb.create_sheet("Statistic Moments")
    write(0, 0, "Sample Name", style="header")
    ws.merge_cells(start_row=1, start_column=1, end_row=2, end_column=1)
    ws.column_dimensions[column_to_char(0)].width = 16
    headers = []
    sub_headers = ["Proportion [%]",
                   "Mean [φ]",
                   "Mean [μm]",
                   "Standard Deviation [φ]",
                   "Standard Deviation [μm]",
                   "Skewness",
                   "Kurtosis"]
    for i in range(result.n_components):
        write(0, i*len(sub_headers)+1, f"C{i+1}", style="header")
        ws.merge_cells(start_row=1, start_column=i*len(sub_headers)+2, end_row=1, end_column=(i+1)*len(sub_headers)+1)
        headers.extend(sub_headers)
    for col, value in enumerate(headers, 1):
        write(1, col, value, style="header")
        ws.column_dimensions[column_to_char(col)].width = 10

    for i, sample in enumerate(result.dataset.samples):
        row = i + 2
        if row % 2 == 0:
            style = "normal_light"
        else:
            style = "normal_dark"
        write(row, 0, sample.name, style=style)
        for j in range(result.n_components):
            true_index = sort_map[j]
            s = get_all_statistic(result.dataset.classes_μm, result.dataset.classes_φ, result.components[i, true_index])
            write(row, true_index*len(sub_headers)+1, result.proportions[i, 0, true_index] * 100, style=style)
            write(row, true_index*len(sub_headers)+2, s["logarithmic"]["mean"], style=style)
            write(row, true_index*len(sub_headers)+3, s["geometric"]["mean"], style=style)
            write(row, true_index*len(sub_headers)+4, s["logarithmic"]["std"], style=style)
            write(row, true_index*len(sub_headers)+5, s["geometric"]["std"], style=style)
            write(row, true_index*len(sub_headers)+6, s["logarithmic"]["skewness"], style=style)
            write(row, true_index*len(sub_headers)+7, s["logarithmic"]["kurtosis"], style=style)
        if progress_callback is not None:
            progress_callback(i / result.n_samples * 0.1 + 0.2)

    logger.debug("Creating the `Unmixed Components` sheet.")
    ws = wb.create_sheet("Unmixed Components")
    ws.merge_cells(start_row=1, start_column=1, end_row=1, end_column=2)
    write(0, 0, "Sample Name", style="header")
    ws.column_dimensions[column_to_char(0)].width = 16
    for col, value in enumerate(result.dataset.classes_μm, 2):
        write(0, col, value, style="header")
        ws.column_dimensions[column_to_char(col)].width = 10

    predict = result.proportions @ result.components
    row = 1
    for i, sample in enumerate(result.dataset.samples):
        if i % 2 == 0:
            style = "normal_light"
        else:
            style = "normal_dark"
        write(row, 0, sample.name, style=style)
        ws.merge_cells(start_row=row+1, start_column=1, end_row=row+result.n_components+1, end_column=1)
        for j in range(result.n_components):
            write(row, 1, f"C{j+1}", style=style)
            true_index = sort_map[j]
            for col, value in enumerate(result.components[i, true_index]*result.proportions[i, 0, true_index], 2):
                write(row, col, value, style=style)
            row += 1
        write(row, 1, "Sum", style=style)
        for col, value in enumerate(predict[i, 0], 2):
            write(row, col, value, style=style)
        row += 1
        if progress_callback is not None:
            progress_callback(i / result.n_samples * 0.2 + 0.3)

    logger.debug("Creating separate sheets for all components.")
    for j in range(result.n_components):
        ws = wb.create_sheet(f"C{j+1}")
        write(0, 0, "Sample Name", style="header")
        ws.column_dimensions[column_to_char(0)].width = 16
        for col, value in enumerate(result.dataset.classes_μm, 1):
            write(0, col, value, style="header")
            ws.column_dimensions[column_to_char(col)].width = 10
        true_index = sort_map[j]
        for i, sample in enumerate(result.dataset.samples):
            row = i + 1
            if row % 2 == 0:
                style = "normal_dark"
            else:
                style = "normal_light"
            write(row, 0, sample.name, style=style)
            for col, value in enumerate(result.components[i, true_index], 1):
                write(row, col, value, style=style)
            if progress_callback is not None:
                progress_callback(((i / result.n_samples) + j) / result.n_components * 0.5 + 0.5)

    wb.save(filename)
    wb.close()
    if progress_callback is not None:
        progress_callback(1.0)
    logger.info(f"The UDM result have been saved to the Excel file: [{filename}].")



__all__ = [
    "column_to_char",
    "to_cell_name",
    "prepare_styles"]

import openpyxl
from openpyxl.styles import (Alignment, Border, Font, NamedStyle, PatternFill,
                             Side)


def column_to_char(column_index: int):
    column = column_index + 1
    column_str = str()
    while column != 0:
        res = column % 26
        if res == 0:
            res = 26
            column -= 26
        column_str = chr(ord('A') + res - 1) + column_str
        column = column // 26
    return column_str

def to_cell_name(row: int, column: int):
    return "{0}{1}".format(column_to_char(column), row+1)

def prepare_styles(wb: openpyxl.Workbook):
    normal_light_style = NamedStyle(name="normal_light")
    normal_light_style.font = Font(size=12, name="Arial", color="000000")
    normal_light_style.alignment = Alignment(
        horizontal='center',
        vertical='center',
        text_rotation=0,
        wrap_text=False,
        shrink_to_fit=False,
        indent=0)
    normal_light_style.fill = PatternFill(
        patternType="solid",
        fgColor="c4cbcf")
    wb.add_named_style(normal_light_style)

    normal_dark_style = NamedStyle(name="normal_dark")
    normal_dark_style.font = Font(size=12, name="Arial", color="000000")
    normal_dark_style.alignment = Alignment(
        horizontal='center',
        vertical='center',
        text_rotation=0,
        wrap_text=False,
        shrink_to_fit=False,
        indent=0)
    normal_dark_style.fill = PatternFill(
        patternType="solid",
        fgColor="b2bbbe")
    wb.add_named_style(normal_dark_style)

    warning_style = NamedStyle(name="warning")
    warning_style.font = Font(size=12, name="Arial", color="ffffff")
    warning_style.alignment = Alignment(
        horizontal='center',
        vertical='center',
        text_rotation=0,
        wrap_text=False,
        shrink_to_fit=False,
        indent=0)
    warning_style.fill = PatternFill(
        patternType="solid",
        fgColor="ff9900")
    wb.add_named_style(warning_style)

    header_style = NamedStyle(name="header")
    header_style.font = Font(size=14, bold=True, name="Times New Roman", color="ffffff")
    header_style.alignment = Alignment(
        horizontal='center',
        vertical='center',
        text_rotation=0,
        wrap_text=True,
        shrink_to_fit=False,
        indent=0)
    header_style.fill = PatternFill(
        patternType="solid",
        fgColor="475164")
    wb.add_named_style(header_style)

    description_style = NamedStyle(name="description")
    description_style.font = Font(size=14, bold=True, name="Times New Roman", color="ffffff")
    description_style.alignment = Alignment(
        horizontal='left',
        vertical='center',
        text_rotation=0,
        wrap_text=False,
        shrink_to_fit=False,
        indent=0)
    description_style.fill = PatternFill(
        patternType="solid",
        fgColor="475164")
    wb.add_named_style(description_style)



from __future__ import annotations

import copy
import typing

import numpy as np

from ..model import GrainSizeDataset
from ..ssu import get_distance_function
from ._kernel import KernelType
from ._setting import EMMAAlgorithmSetting


class EMMAResult:
    def __init__(self, dataset: GrainSizeDataset,
                 kernel_type: KernelType,
                 n_members: int,
                 initial_parameters: np.ndarray,
                 resolver_setting: EMMAAlgorithmSetting,
                 proportions: np.ndarray,
                 end_members: np.ndarray,
                 time_spent: float,
                 history: typing.List[typing.Tuple[np.ndarray, np.ndarray]]):
        self.__dataset = dataset
        self.__kernel_type = kernel_type
        self.__n_members = n_members
        self.__initial_parameters = initial_parameters
        self.__resolver_setting = resolver_setting
        self.__proportions = proportions
        self.__end_members = end_members
        self.__time_spent = time_spent
        self.__history = history

    @property
    def dataset(self) -> GrainSizeDataset:
        return self.__dataset

    @property
    def n_samples(self) -> int:
        return self.dataset.n_samples

    @property
    def n_classes(self) -> int:
        return len(self.__dataset.classes_φ)

    @property
    def kernel_type(self) -> KernelType:
        return self.__kernel_type

    @property
    def n_members(self) -> int:
        return self.__n_members

    @property
    def initial_parameters(self) -> int:
        return self.__initial_parameters

    @property
    def resolver_setting(self) -> EMMAAlgorithmSetting:
        return self.__resolver_setting

    @property
    def proportions(self) -> np.ndarray:
        return self.__proportions

    @property
    def end_members(self) -> np.ndarray:
        return self.__end_members

    @property
    def time_spent(self) -> float:
        return self.__time_spent

    @property
    def n_iterations(self) -> int:
        return len(self.__history)

    @property
    def history(self) -> typing.Iterable[EMMAResult]:
        for fractions, end_members in self.__history:
            copy_result = copy.copy(self)
            copy_result.__proportions = fractions
            copy_result.__end_members = end_members
            yield copy_result

    def get_distance(self, distance: str) -> float:
        distance_func = get_distance_function(distance)
        predict = self.__proportions @ self.__end_members
        return distance_func(predict, self.__dataset.distribution_matrix)

    def get_distance_series(self, distance: str) -> np.ndarray:
        distances = []
        distance_func = get_distance_function(distance)
        for fractions, end_members in self.__history:
            predict = fractions @ end_members
            distance = distance_func(predict, self.__dataset.distribution_matrix)
            distances.append(distance)
        distances = np.array(distances)
        return distances

    def get_class_wise_distances(self, distance: str) -> np.ndarray:
        distance_func = get_distance_function(distance)
        predict = self.__proportions @ self.__end_members
        distances = distance_func(predict, self.__dataset.distribution_matrix, axis=0)
        return distances

    def get_sample_wise_distances(self, distance: str) -> np.ndarray:
        distance_func = get_distance_function(distance)
        predict = self.__proportions @ self.__end_members
        distances = distance_func(predict, self.__dataset.distribution_matrix, axis=1)
        return distances
